# -*- coding: utf-8 -*-

"""Command for running CrunchyFrog from local directory."""

import sys

from distutils.core import Command
from pkg_resources import load_entry_point


class run(Command):

    description = 'Launch CrunchyFrog from the local directory'
    user_options = [
        ('appargs=', None, 'Application arguments'),
        ]
    sub_commands = [('build', None)]

    def initialize_options(self):
        self.appargs = ''

    def finalize_options(self):
        pass

    def run(self):
        for cmd_name in self.get_sub_commands():
            self.run_command(cmd_name)

        del sys.argv[1:]
        new_argv = [x for x in self.appargs.split(' ') if x]
        starter = load_entry_point(('crunchyfrog==%s'
                                    % (self.distribution.get_version(),)),
                                   'gui_scripts', 'crunchyfrog')
        starter(['crunchyfrog']+new_argv)



# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""SQL-specific TextView

This module provides a SQL-specific TextView widget. Depending on
your system it is a gtksourceview2 or gtksourceview implementation.

Both implementations inherit ``SQLViewBase`` which connects to the
configuration instance to update some options when they are changed
through the preferences dialog.

Example usage
=============

    .. sourcecode:: python

        >>> from cf.ui.widgets.sqlview import SQLView
        >>> import gtk
        >>> w = gtk.Window()
        >>> sw = gtk.ScrolledWindow()
        >>> view = SQLView(app)
        >>> view.get_buffer().set_text("select * from foo")
        >>> sw.add(view)
        >>> w.add(sw)
        >>> w.show_all()

"""

# The code for highlighting text marks is derived from the browse view
# in Giggle (http://live.gnome.org/giggle) written by
# Mathias Hasselmann <mathias@taschenorakel.de>.

import cairo
import gtk
import gtksourceview2
import gobject
import pango

try:
    import gconf
    HAVE_GCONF = True
except ImportError:
    HAVE_GCONF = False

import sqlparse


class SQLView(gtksourceview2.View):
    """SQLViewBase implementation"""

    def __init__(self, win, editor=None):
        gtksourceview2.View.__init__(self)
        gtk.widget_push_composite_child()
        self.editor = editor
        self.buffer = gtksourceview2.Buffer()
        self.set_buffer(self.buffer)
        self.set_show_line_marks(True)
        width, height = gtk.icon_size_lookup(gtk.ICON_SIZE_MENU)
        pixbuf = gtk.gdk.Pixbuf(gtk.gdk.COLORSPACE_RGB, True,
                                8, width, height)
        pixbuf.fill(0)
        self.set_mark_category_pixbuf('sql-start', pixbuf)
        self.set_mark_category_pixbuf('sql-end', pixbuf)
        lang_manager = gtksourceview2.language_manager_get_default()
        self.buffer.set_language(lang_manager.get_language('sql'))
        self.app = win.app
        self._sig_app_config_changed = self.app.config.connect(
            'changed', self.on_config_changed)
        self.update_textview_options()
        self._buffer_changed_cb = None
        self.buffer.connect('end-user-action', self.on_buffer_changed)
        self.buffer.connect('mark-set', self.on_mark_set)
        if self.editor is not None:
            self.editor.connect('connection-changed', lambda e, c:
                                self.on_buffer_changed(self.buffer))
        self.connect('expose-event', self.on_expose)
        self._sql_marks = set()
        self.connect('destroy', self.on_destroy)

    def on_buffer_changed(self, buffer):
        """Installs timeout callback to self.buffer_changed_cb()."""
        if self._buffer_changed_cb is not None:
            gobject.source_remove(self._buffer_changed_cb)
            self._buffer_changed_cb = None
        self._buffer_change_cb = gobject.idle_add(self.buffer_changed_cb,
                                                  buffer)

    def on_config_changed(self, config, option, value):
        """Updates view and buffer on configuration change."""
        if option.startswith('editor.'):
            gobject.idle_add(self.update_textview_options)
        if option == "sqlparse.enabled":
            gobject.idle_add(self.buffer_changed_cb, self.buffer)

    def on_destroy(self, *args):
        self.app.config.disconnect(self._sig_app_config_changed)

    def on_expose(self, view, event):
        left_margin = view.get_window(gtk.TEXT_WINDOW_LEFT)

        if left_margin != event.window:
            return False

        buffer_ = view.get_buffer()
        color = self.get_style().base[gtk.STATE_SELECTED]

        cr = event.window.cairo_create()
        cr.rectangle(event.area.x, event.area.y,
                     event.area.width, event.area.height)
        cr.clip()

        visible_rect = view.get_visible_rect()
        iter_ = view.get_iter_at_location(visible_rect.x, visible_rect.y)
        y, _ = view.get_line_yrange(iter_)
        offset = y-visible_rect.y
        visible_rect.y += visible_rect.height

        margin_width, _ = left_margin.get_size()
        cr.translate(margin_width-16, offset)

        forward = True
        within_statement = self._iter_in_statement(iter_.get_line())
        curr = self.get_current_statement()
        if curr:
            clstart, clend = [x.get_line() for x in curr]
        else:
            clstart = clend = None
        current = False

        while not iter_.is_end():
            y, height = view.get_line_yrange(iter_)
            if y >= visible_rect.y:
                break
            lineno = iter_.get_line()
            start = False
            end = False
            for mark in buffer_.get_source_marks_at_line(lineno, None):
                name = mark.get_category()
                if name.startswith('sql'):
                    if name.endswith('-start'):
                        within_statement = True
                        start = True
                    elif name.endswith('-end'):
                        end = True
            if clstart is not None and clend is not None:
                current = (lineno>=clstart and lineno<=clend)
            else:
                current = False
            if within_statement:
                self._render_sql_marker(cr, 16, height, color,
                                        start, end, current)
            if end == True:
                within_statement = False
                current = False
            cr.translate(0, height)
            forward = iter_.forward_line()

    def _render_sql_marker(self, cr, width, height, color,
                           start, end, current):
        if current:
            alpha = 1
        else:
            alpha = 0.5

        x0 = 2
        y0 = 0
        x1 = width - 3
        y1 = height - 1

        r = color.red / 65535.0
        g = color.green / 65535.0
        b = color.blue / 65535.0

        gradient = cairo.LinearGradient(0, 0, 0, height)

        if start:
            gradient.add_color_stop_rgba(0.0, r, g, b, 0.0 * alpha)
            gradient.add_color_stop_rgba(0.1, r, g, b, 0.3 * alpha)
            gradient.add_color_stop_rgba(0.4, r, g, b, 0.8 * alpha)
            cr.move_to(x0, y0 + 9)
            cr.curve_to(x0, y0 + 1, x0, y0 + 1, x0 + 8, y0 + 1)
            cr.line_to(x1 - 8, y0 + 1)
            cr.curve_to(x1, y0 + 1, x1, y0 + 1, x1, y0 + 9)
	else:
            gradient.add_color_stop_rgba(0.0, r, g, b, 0.6 * alpha)
            cr.move_to(x0, y0)
            cr.line_to(x1, y0)

	if end:
            gradient.add_color_stop_rgba(1.0, r, g, b, 0.0 * alpha)
            gradient.add_color_stop_rgba(0.9, r, g, b, 0.3 * alpha)
            gradient.add_color_stop_rgba(0.6, r, g, b, 0.6 * alpha)
            cr.line_to(x1, y1 - 9)
            cr.curve_to(x1, y1 - 1, x1, y1 - 1, x1 - 8, y1 - 1)
            cr.line_to(x0 + 8, y1 - 1)
            cr.curve_to(x0, y1 - 1, x0, y1 - 1, x0, y1 - 9)
	else:
            gradient.add_color_stop_rgba(1.0, r, g, b, 0.6 * alpha)
            cr.line_to(x1, y1 + 1)
            cr.line_to(x0, y1 + 1)

        cr.set_source(gradient)
	cr.fill()

	x0 += 0.5
	y0 += 0.5
	x1 += 0.5
	y1 += 0.5

        if start:
            if end:
                cr.move_to(x0, y0 + 9)
            else:
                cr.move_to(x0, y1)
                cr.line_to(x0, y0 + 8)

            cr.curve_to(x0, y0 + 1, x0, y0 + 1, x0 + 8, y0 + 1)
            cr.line_to(x1 - 8, y0 + 1)
            cr.curve_to(x1, y0 + 1, x1, y0 + 1, x1, y0 + 9)
	else:
            cr.move_to(x1, y0)

	if end:
            cr.line_to(x1, y1 - 9)
            cr.curve_to(x1, y1 - 1, x1, y1 - 1, x1 - 8, y1 - 1)
            cr.line_to(x0 + 8, y1 - 1)
            cr.curve_to(x0, y1 - 1, x0, y1 - 1, x0, y1 - 9)
	else:
            cr.line_to(x1, y1 + 1)
            cr.move_to(x0, y1)

	if not start:
            cr.line_to(x0, y0)

	cr.set_line_width(1)
	cr.set_source_rgba(r, g, b, alpha)
	cr.stroke()

    def on_mark_set(self, buffer_, iter_, mark):
        insert = buffer_.get_insert()
        if insert != mark:
            return False
        self.queue_draw()

    def _iter_in_statement(self, lineno):
        for start, end in self.get_statements():
            startl = start.get_line()
            endl = end.get_line()
            if lineno >= startl and lineno <= endl:
                return True
        return False

    def buffer_changed_cb(self, buffer):
        """Update marks."""
        # TODO(andi): Needs optimizations. Is there a way to fetch all marks?
        while self._sql_marks:
            mark = self._sql_marks.pop()
            self.buffer.delete_mark(mark)
        cursor_iter = buffer.get_iter_at_mark(buffer.get_insert())
        cursor_line = cursor_iter.get_line()
        if self.app.config.get("sqlparse.enabled", True):
            for iter_start, iter_end in self.find_statements():
                # Mark beginning of line
                iter_start.set_line_offset(0)
                iter_end.set_line_offset(0)
                mark = buffer.create_source_mark(None, 'sql-start',
                                                 iter_start)
                self._sql_marks.add(mark)
                mark = buffer.create_source_mark(None, 'sql-end', iter_end)
                self._sql_marks.add(mark)
        self._buffer_changed_cb = None
        self.queue_draw()
        return False

    def update_textview_options(self):
        c = self.app.config
        buf = self.get_buffer()
        if c.get('editor.wrap_text'):
            if c.get('editor.wrap_split'):
                self.set_wrap_mode(gtk.WRAP_CHAR)
            else:
                self.set_wrap_mode(gtk.WRAP_WORD)
        else:
            self.set_wrap_mode(gtk.WRAP_NONE)
        self.set_show_line_numbers(c.get('editor.display_line_numbers'))
        self.set_highlight_current_line(c.get('editor.highlight_current_line'))
        self.set_insert_spaces_instead_of_tabs(c.get('editor.insert_spaces'))
        self.set_auto_indent(c.get('editor.auto_indent'))
        if c.get('editor.default_font'):
            if HAVE_GCONF:
                client = gconf.client_get_default()
                default_font = '/desktop/gnome/interface/monospace_font_name'
                font = client.get_string(default_font)
            else:
                font = 'Monospace 10'
        else:
            font = c.get('editor.font')
        self.modify_font(pango.FontDescription(font))
        self.set_show_right_margin(c.get('editor.right_margin'))
        self.set_right_margin_position(c.get('editor.right_margin_position'))
        buf.set_highlight_matching_brackets(c.get('editor.bracket_matching'))
        self.set_tab_width(c.get('editor.tabs_width'))
        sm = gtksourceview2.style_scheme_manager_get_default()
        scheme = sm.get_scheme(c.get('editor.scheme'))
        buf.set_style_scheme(scheme)

    def find_statements(self):
        """Finds statements in current buffer.

        Returns:
            List of 2-tuples (start iter, end iter).
        """
        buffer_ = self.get_buffer()
        buffer_start, buffer_end = buffer_.get_bounds()
        content = buffer_.get_text(buffer_start, buffer_end)
        iter_ = buffer_.get_start_iter()
        for stmt in sqlparse.split(content):
            if not stmt.strip():
                continue
            # FIXME: Does not work if linebreaks in buffers are not '\n'.
            #    sqlparse unifies them!
            bounds = iter_.forward_search(stmt.strip(),
                                          gtk.TEXT_SEARCH_TEXT_ONLY)
            if bounds is None:
                continue
            start, end = bounds
            yield start, end
            iter_ = end
            if not iter_:
                raise StopIteration
        raise StopIteration

    def get_statements(self):
        """Returns iter 2-tuples for marked statements."""
        buffer_ = self.buffer
        iter_ = buffer_.get_iter_at_line(0)
        start = None
        end = None
        while not iter_.is_end():
            lineno = iter_.get_line()
            for mark in buffer_.get_source_marks_at_line(lineno, None):
                name = mark.get_category()
                if name.endswith('-start'):
                    start = buffer_.get_iter_at_mark(mark)
                elif name.endswith('-end'):
                    end = buffer_.get_iter_at_mark(mark)
            if start is not None and end is not None:
                end.forward_to_line_end()
                yield start, end
                start = end = None
            iter_.forward_line()
        raise StopIteration

    def get_current_statement(self):
        """Returns iters for statement where insert mark is or None."""
        iter_ = self.buffer.get_iter_at_mark(self.buffer.get_insert())
        lineno = iter_.get_line()
        for start, end in self.get_statements():
            startl = start.get_line()
            endl = end.get_line()
            if lineno >= startl and lineno <= endl:
                return start, end
        return None



# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Application object"""

import logging
import os
import sqlite3
import sys
import threading
import urlparse
import webbrowser

import gobject
import gtk

from cf import release, USER_DIR, MANUAL_URL, DATA_DIR, USER_CONFIG_DIR
from config import Config
from plugins.core import PluginManager
from cf.db import DatasourceManager
from cf.ui import dialogs
from cf.ui.datasources import DatasourcesDialog
from cf.ui.widgets import ConnectionsDialog
from ui.mainwindow import MainWindow
from ui.prefs import PreferencesDialog
from cf.userdb import UserDB
from cf import autocompletion

import logging


class CFApplication(gobject.GObject):
    """Main application object.

    An instance of this class is accessible in almost every class in
    CrunchyFrog through the app attribute. It provides access
    to application internals.

    Instance attributes:
      config: Configuration
      userdb: User database
      plugins: Plugin registry
      datasources: Datasource information
    """

    def __init__(self, options):
        """Constructor.

        Parameter:
          options: Command line options as returned by optparse.
        """
        self.__gobject_init__()
        self._instances = []
        self.__icount = 0
        self.options = options
        self.icon_theme = gtk.icon_theme_get_default()

    def init(self):
        """Initializes the application"""
        self.cb = CFAppCallbacks()
        self.__shutdown_tasks = []
        self.config = Config(self, self.options.config)
        self.userdb = UserDB(self)
        try:
            self._check_version()
        except:
            logging.exception('Exception during version update:')
        self.run_listener()
        self.plugins = PluginManager(self)
        self.datasources = DatasourceManager(self)
        self.recent_manager = gtk.recent_manager_get_default()
        autocompletion.setup(self)

    def _check_version(self):
        """Run possible version updates."""
        version_file = os.path.join(USER_CONFIG_DIR, 'VERSION')
        if os.path.isfile(version_file):
            dir_version = open(version_file).read()
        else:
            # Upgrade to 0.4.0 regression
            version_file = os.path.expanduser('~/.crunchyfrog/VERSION')
            if os.path.isfile(version_file):
                dir_version = open(version_file).read()
            else:
                dir_version = None
            version_file = os.path.join(USER_CONFIG_DIR, 'VERSION')
        if dir_version is not None and dir_version != release.version:
            self._move_user_files(dir_version, release.version)
            self._datasources_db2url(dir_version, release.version)
            # Do upgrades here
            f = open(version_file, "w")
            f.write(release.version)
            f.close()

    def _move_user_files(self, dir_version, release):
        """Move user specific files and directories to new locations."""
        # Upgrade to 0.4.0: User specific files now life in proper locations
        #   following XDG Base Directory specification.
        import shutil
        import cf
        old_user_dir = os.path.expanduser('~/.crunchyfrog')
        logging.info('Migrating user dir %s', old_user_dir)
        if not os.path.exists(old_user_dir):
            return
        for fname in ['user.db', 'VERSION']:
            path = os.path.join(old_user_dir, fname)
            if not os.path.exists(path):
                continue
            shutil.move(path,
                        os.path.join(cf.USER_CONFIG_DIR, fname))
        if os.path.exists(os.path.join(old_user_dir, 'plugins')):
            shutil.move(os.path.join(old_user_dir, 'plugins'),
                        os.path.join(cf.USER_DIR, 'plugins'))
        shutil.rmtree(old_user_dir)

    def _datasources_db2url(self, dir_version, release):
        """Move data source config from userdb to datasources.cfg."""
        # Upgrade to 0.4.0: Data source are no longer stored in userdb.
        # TODO: Remove this hook in some upcoming major version.
        import cPickle
        from cf.db import Datasource
        from cf.db.url import URL
        from cf.ui import dialogs
        manager = DatasourceManager(self)
        x = self.userdb.get_table_version('datasource')
        if x is None:  # already dropped
            return
        sql = ('select name, description, backend, options, password '
               'from datasource')
        attribs = ['database', 'host', 'port', 'user']
        try:
            self.userdb.cursor.execute(sql)
            attribs.append('password')
        except sqlite3.OperationalError:  # password column is already gone
            sql = ('select name, description, backend, options '
                   'from datasource')
            self.userdb.cursor.execute(sql)
        ldap_found = False
        for item in self.userdb.cursor.fetchall():
            old_backend = item[2].split('.')[-1]
            opts = cPickle.loads(str(item[3]))
            if old_backend in ('mssql', 'oracle', 'mysql', 'postgres'):
                url = URL(old_backend)
                for name in attribs:
                    if name in opts:
                        setattr(url, name, opts[name])
                if len(item) == 5:
                    url.password = item[4]
                ds = Datasource(manager)
                ds.url = url
                ds.name = item[0]
                ds.decription = item[1]
                ds.ask_for_password = opts.get('ask_for_password', False)
                manager.save(ds)
            elif old_backend == 'sqlite':
                url = URL(old_backend)
                url.database = opts.get('filename', None)
                ds = Datasource(manager)
                ds.url = url
                ds.name = item[0]
                ds.description = item[1]
                ds.ask_for_password = opts.get('ask_for_password', False)
                manager.save(ds)
            elif old_backend == 'ldap':
                ldap_found = True
        if ldap_found:
            dialogs.warning('Warning',
                            'Sorry, LDAP server are no longer supported.')
        self.userdb.drop_table('datasource')

    # ---
    # Instance handling
    # ---

    def new_instance(self, args=None, show=True):
        """Creates a new instances.

        Arguments:
          args: List of filenames to open in new instance (default: None).
          show: If True, instance.show() is called here (default: True).

        Returns:
          The new instance.
        """
        create_editor = not args
        instance = MainWindow(self, create_editor=create_editor)
        if args is None:
            args = []
        for arg in args:
            instance.editor_create(arg)
        self._instances.append(instance)
        instance.connect("destroy", self.on_instance_destroyed)
        if show:
            instance.show()
        self.cb.emit("instance-created", instance)
        return instance

    def on_instance_destroyed(self, instance):
        instances = self.get_data("instances")
        if instance in self._instances:
            self._instances.remove(instance)
        if len(self._instances) == 0:
            self.shutdown()

    def get_instance(self, instance_id):
        """Returns an instances identified by ID."""
        for instance in self._instances:
            if id(instance) == instance_id:
                return instance

    def get_instances(self):
        """Returns a list of active instances."""
        return self._instances

    # ---

    def manage_connections(self, win):
        """Displays a dialog to manage connections."""
        dlg = ConnectionsDialog(win)
        dlg.run()
        dlg.destroy()

    def preferences_show(self, win, mode="editor"):
        """Displays the preferences dialog"""
        dlg = PreferencesDialog(win, mode=mode)
        dlg.run() # IGNORE:E1101 - Generic method
        dlg.destroy() # IGNORE:E1101 - Generic method

    def register_shutdown_task(self, func, msg, *args, **kwargs):
        """Registers a new shutdown task.

        Shutdown tasks are executed when the shutdown method is
        called.

        Parameter:
          func: A callable to execute.
          msg:  A human readable message explaining what the task does.
          args, kwargs: arguments and keyword arguments for func (optional).
        """
        self.__shutdown_tasks.append((func, msg, args, kwargs))

    def run_listener(self):
        """Creates and runs a simple socket server."""
        try:
            from cf.ipc import IPCListener
        except AttributeError:
            self.ipc_listener = None
            return
        self.ipc_listener = IPCListener(self)
        t = threading.Thread(target=self.ipc_listener.serve_forever)
        t.setDaemon(True)
        gobject.idle_add(t.start)

    def show_help(self, topic='index'):
        """Opens a webbrowser with a help page.

        Arguments:
          topic: If given, the URL points to the selected topic.
        """
        url = os.path.join(MANUAL_URL, '%s.html' % topic)
        parsed = urlparse.urlparse(url)
        if parsed[0] in ('file', ''):
            local_path = parsed[2]
        else:
            local_path = url
        if not os.path.exists(local_path):
            dialogs.error(_(u"Documentation not found"),
                          _(u"Could not find documentation at %s.") % url)
        else:
            webbrowser.open(url)

    def shutdown(self):
        """Execute all shutdown tasks and quit application

        Shutdown tasks are called within a try-except-block. If an
        exception is raised it will be printed to stdout.
        """
        while self.__shutdown_tasks:
            func, msg, args, kwargs = self.__shutdown_tasks.pop()
            logging.info(msg)
            try:
                func(*args, **kwargs)
            except:
                import traceback; traceback.print_exc()
                logging.error("Task failed: %s" % str(sys.exc_info()[1]))
        if self.ipc_listener is not None:
            self.ipc_listener.shutdown()
        gtk.main_quit()

    # ------------------
    # Resources helpers
    # ------------------

    def get_glade_file(self, filename):
        """Expands filename to full path pointing to Glade file."""
        return os.path.join(DATA_DIR, 'glade', filename)

    def load_icon(self, icon_name, size=None, lookup_method=None):
        """Wrapper for gtk.IconTheme.load_icon that catches GError."""
        if size is None:
            size = gtk.ICON_SIZE_MENU
        if lookup_method is None:
            lookup_method = (gtk.ICON_LOOKUP_FORCE_SVG |
                             gtk.ICON_LOOKUP_USE_BUILTIN |
                             gtk.ICON_LOOKUP_GENERIC_FALLBACK)
        if sys.platform.startswith('win'):
            lookup_method = (gtk.ICON_LOOKUP_NO_SVG |
                             gtk.ICON_LOOKUP_USE_BUILTIN |
                             gtk.ICON_LOOKUP_GENERIC_FALLBACK)
        try:
            return self.icon_theme.load_icon(icon_name, size, lookup_method)
        except gobject.GError, err:
            logging.warning(err)
            return self.icon_theme.load_icon('gtk-missing-image',
                                             size, lookup_method)

    def set_status_message(self, msg, context=1):
        for window in self.get_instances():
            window.statusbar.set_message(msg, context)

    def pop_status_message(self, context=1):
        for window in self.get_instances():
            window.statusbar.pop(context)


class CFAppCallbacks(gobject.GObject):
    """Container for application specific signals.

    :Signals:

    instance-created
      ``def callback(application, instance, user_data1, ...)``

      Called when a new instance was created
    """

    __gsignals__ = {
        "instance-created" : (gobject.SIGNAL_RUN_LAST,
                              gobject.TYPE_NONE,
                              (gobject.TYPE_PYOBJECT,)),
    }



# -*- coding: utf-8 -*-

"""Database objects."""

import os
import re

import gtk
import gobject

_ = lambda x: x

from cf import DATA_DIR


class PropertyAttribute(object):
    """Connects a instance attribute to a property."""

    def __init__(self, prop):
        self._prop = prop

    def __get__(self, instance, owner):
        return instance.get_property(self._prop.name)

    def __set__(self, instance, value):
        instance.set_property(self._prop.name, value)


class DBObjectMeta(gobject.GObjectMeta):

    def __new__(cls, name, bases, dct):
        gproperties = dct.get('__gproperties__', {})
        new_cls = gobject.GObjectMeta.__new__(cls, name, bases, dct)
        # Shift up properties one level
        for prop in new_cls.props:
            if getattr(new_cls, prop.name, None) is not None:
                raise ValueError, 'Invalid property name %s' % prop.name
            setattr(new_cls, prop.name, PropertyAttribute(prop))
        return new_cls



class GObjectBase(gobject.GObject):
    """Base class for all object classes in this module.

    This class provides basic functionality like setting properties and
    creating class attributes from them.
    """

    __gproperties__ = dict()
    __metaclass__ = DBObjectMeta


    def __init__(self, initial_data=None):
        """Inititalize the object.

        If *initial_data* is given, :meth:`apply_data` is called to set
        initial data on instance creation time.
        """
        self.__gobject_init__()
        self._data = {}
        self._set_properties_defaults()
        if initial_data is not None:
            self.apply_data(initial_data)

    def _set_properties_defaults(self):
        """Set default value as given in __gproperties___."""
        [self._data.update({p.name: p.default_value})
         for p in gobject.list_properties(self)]
        [self.set_property(p.name, p.default_value)
         for p in gobject.list_properties(self)]

    def do_get_property(self, property):
        """Property getter.

        Returns:
          The properties values.

        Raises:
          AttributeError if the property doesn't exist.
        """
        if property.name in self._data:
            return self._data[property.name]
        else:
            raise AttributeError, 'unknown property %s' % property.name

    def do_set_property(self, property, value):
        """Property setter.

        Raises:
          AttributeError when a property doesn't exist.
        """
        if self._data.has_key(property.name):
            self._data[property.name] = value
        else:
            raise AttributeError, 'unknown property %s' % property.name

    def apply_data(self, data):
        """Applies data to the object.

        If a key is definied in ``__gproperties__`` the value will be set
        as a property. Otherwise it will be set using the :meth:`set_data`
        method.

        *data* is a dictionary mapping attributes and values.
        """
        for key, value in data.iteritems():
            if key in self._data:
                self.set_property(key, value)
            else:
                self.set_data(key, value)
                setattr(self, key, value)



class DBObject(GObjectBase):

    __gproperties__ = {
        'name' : (
            gobject.TYPE_STRING,
            'Name', 'Object name',
            '',
            gobject.PARAM_READWRITE),
        'comment' : (
            gobject.TYPE_STRING,
            'Comment', 'Object comment',
            '',
            gobject.PARAM_READWRITE),
        'refresh_required' : (
            gobject.TYPE_BOOLEAN,
            'Refresh required', 'Refresh required',
            True,
            gobject.PARAM_READWRITE),
        'has_children': (
            gobject.TYPE_BOOLEAN,
            'Has children', 'Has children',
            True,
            gobject.PARAM_READWRITE),
        }

    typeid = ''
    typestr = u'???'
    icon = 'gtk-missing-image'

    def __init__(self, meta, **kwds):
        GObjectBase.__init__(self, initial_data=kwds)
        self.meta = meta

    def __repr__(self):
        return '<%s "%s" at 0x%07x>' % (self.__class__.__name__,
                                        self.get_display_name(), id(self))

    def get_display_name(self):
        return self.name

    def get_full_name(self):
        return self.name

    def refresh(self):
        self.meta.engine.refresh(self)

    def property_matches(self, name, value):
        if '__' in name:
            name, options = name.split('__', 1)
            if options == 'ilike':
                p = re.compile(value, re.IGNORECASE)
                cmp_func = lambda x, y: bool(p.match(x))
        else:
            cmp_func = lambda x, y: x == y
        if name in self._data:
            cmp_value = self.get_property(name)
        else:
            cmp_value = self.get_data(name)
        return cmp_func(value, cmp_value)

    def has_children(self):
        return self.props.has_children

    def get_children(self):
        return self.meta.get_children(self)

    def get_icon_pixbuf(self):
        """Returns a pixbuf for this object of 16x16 size."""
        if self.icon.startswith('gdbo-'):
            dir_name = os.path.join(DATA_DIR, 'pixmaps')
            fname = os.path.join(dir_name, '%s.png' % self.icon)
            return gtk.gdk.pixbuf_new_from_file(fname)
        else:
            it = gtk.icon_theme_get_default()
            return it.load_icon(self.icon, gtk.ICON_SIZE_MENU,
                                gtk.ICON_LOOKUP_GENERIC_FALLBACK)


class Collection(DBObject):
    """Base class for collections.

    Signals:
      :collection-refreshed: Emited when a collections was refreshed.
        `def callback(collection)`

    @cvar objects: List of objects in this collection.
    @type objects: C{list}
    @cvar typeid: Type identifier
    @type typeid: C{str}
    @cvar typestr: Human-readable representation of C{typeid}
    @type typestr: C{str}
    @cvar objklass: The object class provided by this collection.
    @type objklass: L{DBObject}
    @ivar provider: The current provider.
    @type provider: L{SchemaProvider}
    @ivar backend: The current engine.
    @type backend: L{engines.NullEngine}
    """

    __gproperties__ = {"count" : (gobject.TYPE_LONG,
                                  "Count", "Number of objects as seen in database",
                                  0, long(99999999), 0,
                                  gobject.PARAM_READWRITE),
                       "objects" : (gobject.TYPE_PYOBJECT,
                                    "Objects", "Objects",
                                    gobject.PARAM_READWRITE),
                       }
    __gsignals__ = {"collection-refreshed" : (gobject.SIGNAL_RUN_LAST,
                                         gobject.TYPE_NONE,
                                         tuple())}

    icon = 'gtk-open'

    def __init__(self, meta, items_class, **kwds):
        DBObject.__init__(self, meta, **kwds)
        self.objects = []
        self.meta = meta
        self.items_class = items_class
#        self.provider.connect("backend-changed", self.on_backend_changed)

#    def on_backend_changed(self, provider, backend):
#        """Callback for L{SchemaProvider} C{backend-changed}."""
#        if self.provider == provider and backend != self.backend:
#            self.backend = backend

    def get_display_name(self):
        return self.typestr

#    def check_refresh(self):
#        """
#        Deprecated (see L{filter}).
#        @warning: Deprecated.
#        """
#        deprecated()
#        if self.refresh_required: self.refresh_objects()

    def find(self, *kwds):
        if self.props.refresh_required:
            self.meta.engine.refresh(self)
        res = self.meta.find(cls=self.items_class, parent=self)
        return res

    # FIXME(andi): Replace this with normal list behavior.
    def exists(self, obj):
        """Returns C{True} if C{obj} exists in this collection.

        @param obj: An object
        @type obj: L{DBObject}
        @return: C{True}/C{False}
        @rtype: C{bool}
        """
        return obj in self.objects

    def refresh(self):
        """
        Deprecated (see L{filter}).
        @warning: Deprecated: Use L{filter} instead.
        @todo 0.1.5: Remove this method.
        """
        # deprecated: Use filter() instead.
        self.provider.log("%r.refresh() called")
        self.add_task(self.refresh_count)
        self.add_task(self.refresh_objects)

    def refresh_count(self):
        """
        Deprecated (see L{filter}).
        @warning: Deprecated. Use L{filter} instead.
        @todo 0.1.5: Remove this method.
        """
        # deprecated: Use filter() instead.
        if self.backend:
            self.count = self.backend.get_object_count(self.objklass)
        else:
            self.count = 0

    def refresh_objects(self):
        """
        Deprecated (see L{filter}).
        @warning: Deprecated. Use L{filter} instead.
        @todo 0.1.5: Remove this method.
        """
        # deprecated: Use filter() instead.
        self.objects = []
        if self.backend:
            [self.objects.append(s) for s in self.backend.get_objects(self.objklass)]
        self.refresh_required = False
        self.emit("collection-refreshed")

    def refresh_needed(self):
        """
        Deprecated (see L{filter}).
        @warning: Deprecated. Use L{filter} instead.
        @todo 0.1.5: Remove this method.
        """
        # deprecated: Use filter() instead.
        self.refresh_count()
        return len(self.objects or []) <> self.count

    def all(self):
        """
        Deprecated (see L{filter}).
        @warning: Deprecated. Use L{filter} instead.
        @todo 0.1.5: Remove this method.
        """
        # deprecated: Use filter() instead.
        self.check_refresh()
        return self.objects

    def apply_filter(self, items, *args, **kwargs):
        """Filtering process.

        This method applies the given filter keywords to a list
        of objects and returns a filtered list.

        @param items: List of objects
        @type items: C{list}
        @return: Filtered list of objects
        @rtype: C{list}
        """
        def filters_match(obj, *args, **kwargs):
            for key in kwargs.keys():
                st = kwargs.get(key)
                if hasattr(obj, key):
                    cmp_item = getattr(obj, key)
                else:
                    cmp_item = obj.get_data(key)
                if not cmp_item:
                    return False
                if type(st) in [unicode, str]:
                    try:
                        p = re.compile(st)
                        if type(cmp_item) in [unicode, str]:
                            if not p.search(cmp_item):
                                return False
                        elif hasattr(cmp_item, "name"):
                            if not p.search(getattr(cmp_item, "name")):
                                return False
                        else:
                            return False
                    except:
                        return False
                else:
                    if not cmp_item == st:
                        return False
            return True

        ret = []
        for item in items or []:
            if filters_match(item, **kwargs):
                ret.append(item)
        return list(set(ret))

    ## def filter(self, *args, **kwargs):
    ##     """
    ##     Returns a filtered list of objects.

    ##     If this method is called without any keywords, the
    ##     complete list of objects will be returned.

    ##     All keywords, except the special ones mentioned below, are used
    ##     to filter the results. For example, C{filter(name="foo")} would
    ##     return all items in the collection with "Foo" as their object name.

    ##     The keywords are passed to the engines C{filter} method, which does
    ##     the real filtering and schema queries. But only if the C{_cached}
    ##     keyword is C{False}.

    ##     Objects returned by the engine are filtered (again) by L{apply_filter}.
    ##     Finally the objects are sorted within this method.

    ##     If the C{_create} keyword is present and C{True}, an object
    ##     based on L{objklass} is created with the given keywords, if the filtering
    ##     process returns no results.

    ##     @keyword _create: Create object if it doesn't exist and return it (default: C{False}).
    ##     @type _create: C{bool}
    ##     @keyword _cached: Don't query the database, return objects from internal cache (default: C{False}).
    ##     @type _cached: C{bool}
    ##     @return: Filtered list of objects.
    ##     @rtype: C{list}
    ##     """
    ##     if kwargs.has_key("_create"):
    ##         create_object = kwargs.get("_create")
    ##         del kwargs["_create"]
    ##     else:
    ##         create_object = False
    ##     if kwargs.has_key("_cached"):
    ##         cached = kwargs.get("_cached")
    ##         del kwargs["_cached"]
    ##     else:
    ##         cached = False
    ##     if cached:
    ##         items = self.objects
    ##     else:
    ##         items = self.backend.filter(self, *args, **kwargs)
    ##     ret = self.apply_filter(items,
    ##                             *args, **kwargs)
    ##     if not ret and create_object:
    ##         obj = self.objklass(self.provider, initial_data = kwargs)
    ##         self.objects.append(obj)
    ##         ret = [obj]
    ##     self.provider.cache.append(ret)
    ##     ret.sort()
    ##     return ret



class Argument(DBObject):

    __gproperties__ = {"type" : (gobject.TYPE_PYOBJECT,
                                    "Type", "Argument data type",
                                    gobject.PARAM_READWRITE),
                       "mode" : (gobject.TYPE_STRING,
                                        "In-Out-InOut", "In-Out-InOut",
                                        "",
                                        gobject.PARAM_READWRITE),
                       "name" : (gobject.TYPE_STRING,
                                        "Name", "Name",
                                        "",
                                        gobject.PARAM_READWRITE),
                       "sortorder" : (gobject.TYPE_INT,
                                        "Sort order", "Sort order",
                                        -1, 10000, -1,
                                        gobject.PARAM_READWRITE),
                       "parent" : (gobject.TYPE_PYOBJECT,
                                        "Parent object", "Parent object",
                                        gobject.PARAM_READWRITE)}

    def __init__(self, meta, **kwds):
        DBObject.__init__(self, meta, **kwds)
        self.typeid = "argument"
        self.typestr = _(u"Arguments")


class Arguments(Collection):

    def __init__(self, meta, **kwds):
        Collection.__init__(self, meta, Argument, **kwds)
        self.typeid = "arguments"
        self.typestr = _(u"Arguments")


class Cache(Collection):

    def __init__(self, provider):
        self.typeid = "dbschema_cache"
        self.typestr = _(u"dbschema_cache")
        self.backend = None
        Collection.__init__(self, provider, self.backend, None)

    def append(self, items):
        self.objects += items
        self.objects = list(set(self.objects))

class Columns(Collection):

    def __init__(self, meta, **kwds):
        Collection.__init__(self, meta, Column, **kwds)
        self.typeid = "columns"
        self.typestr = _(u"Columns")

    def new(self):
        col = Column(self.meta, self.parent)
        self.objects.append(col)
        return col

class Column(DBObject):

    __gproperties__ = {"sortorder" : (gobject.TYPE_INT,
                                      "Sort order", "Sort order", -1, 10000, -1,
                                      gobject.PARAM_READWRITE),
                       "type" : (gobject.TYPE_PYOBJECT,
                                 "Data type", "Data type",
                                 gobject.PARAM_READWRITE),
                       "default" : (gobject.TYPE_PYOBJECT,
                                    "Default", "Default value",
                                    gobject.PARAM_READWRITE),
                       "pk" : (gobject.TYPE_BOOLEAN,
                               "Primary key", "Primary key", False,
                               gobject.PARAM_READWRITE),
                       "nullable" : (gobject.TYPE_BOOLEAN,
                                     "Nullable", "Nullable", False,
                                     gobject.PARAM_READWRITE),
                       "parent" : (gobject.TYPE_PYOBJECT,
                                   "Parent", "Parent object",
                                   gobject.PARAM_READWRITE)}

    icon = 'gdbo-column'

    def __init__(self, meta, parent, *args, **kwargs):
        kwargs["parent"] = parent
        DBObject.__init__(self, meta, *args, **kwargs)
        self.typeid = "column"
        self.typestr = _(u"Column")

    def __cmp__(self, other):
        if not isinstance(other, Column) and isinstance(other, DBObject):
            return cmp(self.props.name, other.props.name)
        elif isinstance(other, Column):
            return cmp(self.props.sortorder, other.props.sortorder)
        else:
            return 1

class Constraints(Collection):

    def __init__(self, meta, **kwds):
        Collection.__init__(self, meta, Constraint, **kwds)
        self.typeid = "constraints"
        self.typestr = _(u"Constraints")

    def new(self):
        c = Constraint(self.provider, self.parent)
        self.objects.append(c)
        return c

class Constraint(DBObject):

    __gproperties__ = {"type" : (gobject.TYPE_STRING,
                                 "Sort order", "Sort order", None,
                                 gobject.PARAM_READWRITE),
                       "check_expression" : (gobject.TYPE_STRING,
                                             "Check expression", "Check expression", "",
                                             gobject.PARAM_READWRITE),
                       "columns" : (gobject.TYPE_PYOBJECT,
                                    "Primary/unique columns", "Primary key/unique columns",
                                    gobject.PARAM_READWRITE),
                       "fkcolumns" : (gobject.TYPE_PYOBJECT,
                                      "Foreign key columns", "Foreign key columns",
                                      gobject.PARAM_READWRITE),
                       "parent" : (gobject.TYPE_PYOBJECT,
                                   "Parent object", "Parent object",
                                   gobject.PARAM_READWRITE)}

    icon = 'gtk-spell-check'

    def __init__(self, meta, parent, *args, **kwargs):
        kwargs["parent"] = parent
        DBObject.__init__(self, meta, *args, **kwargs)
        self.typeid = "constraint"
        self.typestr = _(u"Constraint")
        self.columns = Columns(self.meta, parent=self)
        self.meta.set_object(self.columns)
        self.fkcolumns = Columns(self.meta, parent=self)
        self.props.has_children = False

class Function(DBObject):

    __gproperties__ = {"arguments" : (gobject.TYPE_PYOBJECT,
                                    "Arguments", "Argument list",
                                    gobject.PARAM_READWRITE)}

    icon = 'gdbo-function'

    def __init__(self, meta, **kwds):
        DBObject.__init__(self, meta, **kwds)
        self.typeid = "function"
        self.typestr = _(u"Function")
        self.arguments = Arguments(self.meta, parent=self)
        self.props.has_children = False


class Functions(Collection):

    def __init__(self, meta, **kwds):
        Collection.__init__(self, meta, Function, **kwds)
        self.typeid = "functions"
        self.typestr = _(u"Functions")


class Table(DBObject):

    __gproperties__ = {"columns" : (gobject.TYPE_PYOBJECT,
                                    "Columns", "Column list",
                                    gobject.PARAM_READWRITE),
                       "constraints" : (gobject.TYPE_PYOBJECT,
                                        "Constraints", "Table constraints",
                                        gobject.PARAM_READWRITE)}

    icon = 'gdbo-table'

    def __init__(self, meta, **kwds):
        DBObject.__init__(self, meta, **kwds)
        self.typeid = "table"
        self.typestr = _(u"Table")
        self.columns = Columns(self.meta, parent=self)
        self.meta.set_object(self.columns)
        self.constraints = Constraints(self.meta, parent=self)
        self.meta.set_object(self.constraints)

    def get_full_name(self):
        if hasattr(self, "schema") and not self.schema.is_default:
            n = "%s.%s" % (self.schema.name, self.name)
        else:
            n = self.name
        return n

    def get_insert_sql(self, col_values = dict()):
        sql = "insert into %s" % self.name
        cols = list()
        vals = list()
        my_columns = list()
        [my_columns.append(x.name) for x in self.columns.filter()]
        for key in col_values.keys():
            if not key in my_columns: continue
            cols.append(key)
            vals.append("%r" % col_values[key])
        sql += " (%s) values" % ", ".join(cols)
        sql += " (%s)" % ", ".join(vals)
        return sql

    def get_row_count(self):
        sql = "select count(*) from %s" % self.get_full_name()
        return self.provider.engine.run_query(sql, True)[0]

    def insert(self, *args, **kwargs):
        sql = self.get_insert_sql(kwargs)
        e = self.provider.engine
        e.begin_modification()
        e.modifications.append(sql)
        ret_code, err_msg = e.end_modification(True)
        return ret_code, err_msg




class Tables(Collection):

    def __init__(self, meta, **kwds):
        Collection.__init__(self, meta, Table, **kwds)
        self.typeid = "tables"
        self.typestr = _(u"Tables")


class Schema(DBObject):

    __gproperties__ = {"is_default" : (gobject.TYPE_BOOLEAN,
                                  "Name", "Object name",
                                  False,
                                  gobject.PARAM_READWRITE)}

    def __init__(self, meta, **kwds):
        DBObject.__init__(self, meta, **kwds)
        self.typeid = "schema"
        self.typestr = _(u"Schema")
        self.is_default = False  # Dummy value for now...


class Schemata(Collection):

    def __init__(self, meta):
        Collection.__init__(self, meta, Schema)
        self.typeid = "schemata"
        self.typestr = _(u"Schemata")
#        self.functions = Functions(provider)
#        self.tables = Tables(provider)
#        self.views = Views(provider)

class Sequence(DBObject):

    # Numbers can *very* large, store them as strings...
    __gproperties__ = {"minValue" : (gobject.TYPE_STRING,
                                  "Min. value", "Min. value",
                                  "",
                                  gobject.PARAM_READWRITE),
                       "maxValue" : (gobject.TYPE_STRING,
                                  "Max. value", "Max. value",
                                  "",
                                  gobject.PARAM_READWRITE),
                       "lastValue" : (gobject.TYPE_STRING,
                                  "Last value", "Last value",
                                  "",
                                  gobject.PARAM_READWRITE),
                       "incrementBy" : (gobject.TYPE_STRING,
                                  "Increment by", "Increment by",
                                  "",
                                  gobject.PARAM_READWRITE),}

    icon = 'gtk-sort-descending'

    def __init__(self, meta, **kwds):
        DBObject.__init__(self, meta, **kwds)
        self.typeid = "sequence"
        self.typestr = _(u"Sequence")
        self.props.has_children = False


class Sequences(Collection):

    def __init__(self, meta, **kwds):
        Collection.__init__(self, meta, Sequence, **kwds)
        self.typeid = "sequences"
        self.typestr = _(u"Sequences")

class User(DBObject):

    __gproperties__ = {"isCurrent" : (gobject.TYPE_BOOLEAN,
                                  "Current user", "Current user",
                                  False,
                                  gobject.PARAM_READWRITE)}

    icon = 'gdbo-user'

    def __init__(self, meta, **kwds):
        DBObject.__init__(self, meta, **kwds)
        self.typeid = "user"
        self.typestr = _(u"User")
        self.has_comment = False


class Users(Collection):

    def __init__(self, meta):
        Collection.__init__(self, meta, User)
        self.typeid = "users"
        self.typestr = _(u"Users")

class Trigger(DBObject):

    __gproperties__ = {"table" : (gobject.TYPE_PYOBJECT,
                                    "Table", "Table",
                                    gobject.PARAM_READWRITE),
                       "schema" : (gobject.TYPE_PYOBJECT,
                                    "Schema", "Table's schema",
                                    gobject.PARAM_READWRITE),
                       "source" : (gobject.TYPE_STRING,
                                    "Source", "Source",
                                    None,
                                    gobject.PARAM_READWRITE),}

    typeid = "trigger"
    typestr = _(u"Trigger")


class Triggers(Collection):

    typeid = "triggers"
    typestr = _(u"Trigger")

    def __init__(self, meta, **kwds):
        Collection.__init__(self, meta, Trigger, **kwds)


class Type(DBObject):

    __gproperties__ = {"variableLength" : (gobject.TYPE_BOOLEAN,
                                    "Variable length", "Variable length",
                                    False,
                                    gobject.PARAM_READWRITE),
                       "isComposite" : (gobject.TYPE_BOOLEAN,
                                        "Composite type", "Composite type",
                                        False,
                                        gobject.PARAM_READWRITE)}

    def __init__(self, provider, *args, **kwargs):
        DBObject.__init__(self, provider, *args, **kwargs)
        self.typeid = "type"
        self.typestr = _(u"Type")


class Types(Collection):

    def __init__(self, provider):
        self.backend = None
        self.typeid = "types"
        self.typestr = _(u"Types")
        Collection.__init__(self, provider, self.backend, Type)

class Views(Collection):

    def __init__(self, meta, **kwds):
        Collection.__init__(self, meta, View, **kwds)
        self.typeid = "views"
        self.typestr = _(u"Views")
#        self.tables = Tables()

class View(DBObject):

    __gproperties__ = {"columns" : (gobject.TYPE_PYOBJECT,
                                    "Columns", "Column list",
                                    gobject.PARAM_READWRITE),
                       "src" : (gobject.TYPE_STRING,
                                        "View source", "View source",
                                        "",
                                        gobject.PARAM_READWRITE)}

    icon = 'gdbo-table'

    def __init__(self, meta, **kwds):
        DBObject.__init__(self, meta, **kwds)
        self.typeid = "view"
        self.typestr = _(u"View")
        self.columns = Columns(self.meta, parent=self)
        self.meta.set_object(self.columns)


class Language(DBObject):

    icon = 'stock_script'

    def __init__(self, meta, **kwds):
        DBObject.__init__(self, meta, **kwds)
        self.typeid = 'language'
        self.typestr = _(u'Language')
        self.props.has_children = False


class Languages(Collection):

    def __init__(self, meta):
        Collection.__init__(self, meta, Language)
        self.typeid = "languages"
        self.typestr = _(u"Languages")


class Index(DBObject):

    icon = 'stock_navigator-indexes'
    typeid = 'index'
    typestr = _(u'Index')

    def __init__(self, meta, **kwds):
        DBObject.__init__(self, meta, **kwds)
        self.props.has_children = False


class Indexes(Collection):

    typeid = 'indexes'
    typestr = _(u'Indexes')

    def __init__(self, meta, **kwds):
        Collection.__init__(self, meta, Index, **kwds)


gobject.type_register(Columns)
gobject.type_register(Column)
gobject.type_register(Constraints)
gobject.type_register(Constraint)
gobject.type_register(Functions)
gobject.type_register(Function)
gobject.type_register(Schemata)
gobject.type_register(Schema)
gobject.type_register(Sequences)
gobject.type_register(Sequence)
gobject.type_register(Tables)
gobject.type_register(Table)
gobject.type_register(Triggers)
gobject.type_register(Trigger)
gobject.type_register(Users)
gobject.type_register(User)
gobject.type_register(Views)
gobject.type_register(View)
gobject.type_register(Languages)
gobject.type_register(Language)
gobject.type_register(Index)
gobject.type_register(Indexes)



# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Builtin plugins"""

import logging

from cf.filter.exportfilter import CSVExportFilter
try:
    import ooolib
except ImportError:
    pass
else:
    from cf.filter.exportfilter import OOCalcExportFilter
try:
    import xlwt
except ImportError, err:
    pass
else:
    from cf.filter.exportfilter import XlsExportFilter

try: from cf.db.backends.mysql import MySQL
except ImportError, err:
    logging.warning('Could not import MySQLBackend: %s', err)
try: from cf.db.backends.oracle import Oracle
except ImportError, err:
    logging.warning('Could not import OracleBackend; %s', err)

from cf.shell import CFShell
from cf.library import SQLLibraryPlugin
from cf.nativeshell import NativeShellPlugin




# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Simple UNIX socket server to handle multiple instances.

A very simple protocol is used:

 - Each request and response is a single line.
 - A request has the format 'cmd:[args]'
 - Response is a pickled Python value.
"""

import cPickle
import errno
import logging
import os
import socket
import SocketServer

import gobject
import gtk

from cf import IPC_SOCKET


class IPCRequestHandler(SocketServer.BaseRequestHandler):
    """RequestHandler: Talks to the application instance."""

    def __init__(self, request, client_address, server):
        self.app = server.app
        SocketServer.BaseRequestHandler.__init__(self, request, client_address,
                                                 server)

    def _respond(self, value):
        """Helper that converts the return value and sends it on the socket."""
        logging.info('Response: %r' % value)
        response = cPickle.dumps(value)
        assert len(response) <= 1024, 'Reponse too large'
        try:
            self.request.send(response)
        except socket.error, err:
            if err.args[0] == 32:  # Broken pipe
                pass
            else:
                raise

    def handle(self):
        """Handles a request."""
        data = self.request.recv(1024)
        logging.info('Received: %r' % data)
        cmd, args = data.split(':', 1)
        if cmd == 'ping':
            self._respond(True)
        elif cmd == 'new-instance':
            args = args.split('|')
            gobject.idle_add(lambda: not self.app.new_instance(args))
            self._respond(-1)
        elif cmd == 'get-instances':
            self._respond([(id(i), i.get_title())
                           for i in self.app.get_instances()])
        elif cmd == 'open-uri':
            instance_id, uri = args.split(':', 1)
            for i in self.app.get_instances():
                if id(i) == int(instance_id):
                    i.editor_create(uri)
                    self._respond(True)
            self._respond(False)
        else:
            self._respond('ERROR: Invalid command %r' % data)


class IPCListener(SocketServer.UnixStreamServer):
    """Simple server that listens on a UNIX socket."""

    def __init__(self, app):
        self.app = app
        self.address_family = socket.AF_UNIX
        self.socket_type = socket.SOCK_STREAM
        if os.path.exists(IPC_SOCKET):
            os.remove(IPC_SOCKET)
        SocketServer.UnixStreamServer.__init__(self, IPC_SOCKET,
                                               IPCRequestHandler)
        self.stay_alive = True

    def serve_forever(self):
        while self.stay_alive:
            self.handle_request()

    def shutdown(self):
        self.stay_alive = False
        try:
            os.remove(IPC_SOCKET)
        except OSError, err:
            if err.errno == errno.ENOENT:  # Socket is already gone.
                pass
            else:
                raise


class IPCClient(object):
    """Client."""

    def __init__(self):
        pass

    def communicate(self, msg):
        """Send message to listener and convert the pickled response."""
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        try:
            sock.connect(IPC_SOCKET)
        except socket.error, err:
            msg = 'Could not connect to IPC server on %r.' % IPC_SOCKET
            logging.info(msg)
            raise
        sock.send(msg)
        retval = sock.recv(1024)
        sock.close()
        try:
            retval = cPickle.loads(retval)
        except EOFError:
            retval = None
        return retval

    def get_instances(self):
        """Ask for active instances."""
        try:
            return self.communicate('get-instances:')
        except socket.error, err:
            return []

    def new_instance(self, args=None):
        """Creates a new instance, optionally with args as files to open."""
        if args is not None:
            args = '|'.join(args)
        else:
            args = ''
        try:
            return self.communicate('new-instance:%s' % args)
        except socket.error, err:
            return None

    def open_uri(self, instance_id, uri):
        """Opens a URI in an existing instance."""
        try:
            return self.communicate('open-uri:%d:%s' % (instance_id, uri))
        except socket.error, err:
            return False

    def ping(self):
        """Just say hello."""
        try:
            return self.communicate('ping:')
        except socket.error, err:
            return False


def get_client():
    """Return a client instance."""
    client = IPCClient()
    return client


def is_alive(client):
    """Tries to connect to a running application."""
    response = client.ping()
    if not response:
        return False
    return True



# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Nifty little helpers"""

import re
import urlparse

import gtk
import gobject


# found here: http://www.mail-archive.com/pygtk@daa.com.au/msg12556.html
def Emit(obj, *args):

    def Do_Emit(obj, sig, *args):
        gtk.gdk.threads_enter()
        obj.emit(sig, *args)
        gtk.gdk.threads_leave()
        return False
    gobject.idle_add(Do_Emit, obj, *args, **{"priority" : gobject.PRIORITY_HIGH})
#END: Emit

def normalize_sql(sql):
    """Removes duplicated whitespaces and line breaks."""
    p = re.compile(r'\s+', re.MULTILINE)
    return p.sub(' ', sql)


def to_uri(filename):
    """Converts a filename to URI. It's ok to pass in a URI here."""
    scheme, netloc, path, params, query, fragment = urlparse.urlparse(filename)
    if not scheme:
        uri = urlparse.urlunparse(('file', netloc, path, params,
                                   query, fragment))
    else:
        uri = filename
    return uri


def profile(func):
    """Decorator to profile a function."""
    def wrapper(*args, **kwds):
        import hotshot
        prof = hotshot.Profile("%s.prof" % func.__name__)
        ret = prof.runcall(func, *args, **kwds)
        prof.close()
        return ret
    return wrapper



#!/usr/bin/env python

import os
import pydoc


HEADER = '=-^""""""'
PACKAGES = ["cf"]
BASE_DIR = os.path.abspath(".")
OUT_DIR = os.path.abspath('docs/devguide/source/api/')


def _build_pkg(name, base=BASE_DIR, output=None, level=0, modname=None):
    if modname is None:
        modname = name
    modules = {}
    fullpath = os.path.join(base, name)
    for dirpath, dirnames, filenames in os.walk(fullpath):
        pyfiles = [x for x in filenames if x.endswith('.py')]
        if not pyfiles:
            continue
        modname = dirpath.replace(base, '').replace('/', '.')[1:]
        fname = os.path.join(dirpath, '__init__.py')
        if not os.path.exists(fname):   # Package data
            continue
        s = pydoc.source_synopsis(open(fname)) or "[undocumented]"
        modules[modname] = s.strip()
        for pyfile in pyfiles:
            if pyfile == '__init__.py':
                continue
            xmodname = '.'.join([modname, pyfile.replace('.py', '')])
            fname = os.path.join(dirpath, pyfile)
            s = pydoc.source_synopsis(open(fname)) or "[undocumented]"
            modules[xmodname] = s.strip()
    keys = modules.keys()
    keys.sort()
    for k in keys:
        output = []
        header = ':mod:`%s` -- %s' % (k, modules[k])
        output.append(header)
        output.append(HEADER[k.count('.')]*len(header))
        output.append('')
        output.append('.. automodule:: %s' % k)
        output.append('   :members:')
        output.append('   :synopsis: %s' % modules[k])
        output.append('')
        f = open(os.path.join(OUT_DIR, '%s.rst' % k.replace('.', '_')), 'w')
        f.write('\n'.join(output))
        f.close()

def main():
    for pkg in PACKAGES:
        fname = os.path.join(BASE_DIR, pkg, '__init__.py')
        synopsis = pydoc.source_synopsis(open(fname)) or ''
        header = ':mod:`%s` -- %s' % (pkg, synopsis.strip())
        output = ['*'*len(header), header, '*'*len(header), '']
        output.append('.. toctree::')
        output.append('   :glob:')
        output.append('')
        output.append('   %s_*' % pkg)
        _build_pkg(pkg, output=output, level=0)
        # We've just one package, so we write it out as index.rst.
        # Replace this to '%s.rst' % pkg if we've more than one... ;-)
        f = open(os.path.join(OUT_DIR, 'index.rst'), 'w')
        f.write('\n'.join(output))
        f.close()


if __name__ == "__main__":
    main()




# -*- coding: utf-8 -*-

"""build_api command -- Generate API docs from sources"""

from distutils.command.build import build
from distutils.core import Command
import os

from utils import genapi


class build_api(Command):

    description = 'Generate API docs from sources.'

    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        genapi.main()

build.sub_commands.append(('build_api', None))



# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""SQL library

This module contains the SQL library plugin.

:ToDo's:
    - Re-order categories and statements
    - Double-click / open in editor (drag'n'drop)
"""

from gettext import gettext as _

import gtk
import gobject
import pango

from cf import utils
from cf.plugins.core import BottomPanePlugin
from cf.plugins.mixins import InstanceMixin, UserDBMixin
from cf.ui.pane import PaneItem


class Category(object):
    """Data object holding a category."""

    def __init__(self, id_, name, parent=None):
        self.id = id_
        self.name = name
        self.parent = parent

    def get_update_statement(self):
        """Returns an update SQL."""
        sql = ("update sqllib_cat set name = ?, parent = ? where id = ?")
        return sql, (self.name, self.parent, self.id)

    def get_insert_statement(self):
        """Returns an insert SQL."""
        sql = "insert into sqllib_cat (name, parent) values (?, ?)"
        return sql, (self.name, self.parent )

    def get_delete_statement(self):
        """Returns a delete SQL."""
        sql = 'delete from sqllib_cat where id = ?'
        return sql, (self.id,)


class Statement(object):
    """Data object holding a statement."""

    def __init__(self, id_, title, statement, description=None, category=None):
        self.id = id_
        if title is None:
            title = utils.normalize_sql(statement)
        self.title = title
        self.statement = statement
        self.description = description
        self.category = category

    def get_update_statement(self):
        """Returns an update SQL."""
        sql = ("update sqllib_sql set title = ?, statement = ?, "
               "description = ?, category = ? where id = ?")
        return sql, (self.title or self.statement, self.statement,
                     self.description or None, self.category or None,
                     self.id)

    def get_insert_statement(self):
        """Returns an insert SQL."""
        sql = ("insert into sqllib_sql (title, statement, description, "
               "category) values (?, ?, ?, ?)")
        return sql, (self.title or self.statement, self.statement,
                     self.description or None, self.category or None)

    def get_delete_statement(self):
        """Returns a delete SQL."""
        sql = 'delete from sqllib_sql where id = ?'
        return sql, (self.id,)


class SQLLibraryPlugin(BottomPanePlugin, InstanceMixin, UserDBMixin):
    """SQL library plugin"""

    id = "crunchyfrog.plugin.library"
    name = _(u"Library")
    description = _(u"A personal SQL library")
    icon = "stock_book_green"
    author = "Andi Albrecht"
    license = "GPL"
    homepage = "http://cf.andialbrecht.de"
    version = "0.1"

    def __init__(self, app):
        BottomPanePlugin.__init__(self, app)
        self._views = {}

    def on_save_to_library(self, instance):
        editor = instance.get_active_editor()
        if editor is None:
            return
        statement = editor.get_text()
        entry = Statement(None, None, statement)
        self.save_entry(entry)

    def init_instance(self, instance):
        """Initialize instance."""
        view = SQLLibraryView(self.app, instance, self)
        view.refresh()
        instance.side_pane.add_item(view)
        group = None
        for group in instance.ui.get_action_groups():
            if group.get_name() == 'editor':
                break
        assert group is not None
        entries = (
            ('lib-save', None,
             _(u'Save to library'), None,
             _(u'Save content of the current editor to SQL library.'),
             lambda menuitem: self.on_save_to_library(instance)),
        )
        group.add_actions(entries)
        merge_id = instance.ui.add_ui_from_string(UI)
        self._views[instance] = (view, merge_id)

    def deactivate_instance(self, instance):
        """De-activates an view."""
        if instance in self._views:
            view, merge_id = self._views.pop(instance)
            instance.ui.remove_ui(merge_id)
            for group in instance.ui.get_action_groups():
                if group.get_name() == 'editor':
                    action = group.get_action('lib-save')
                    if action is not None:
                        group.remove_action(action)
                        break
            view.destroy()

    def shutdown(self):
        """Shutdown plugin."""
        while self._views:
            self.deactivate_instance(self._views.keys()[0])

    def userdb_init(self):
        """Initialize user db."""
        if self.userdb.get_table_version("sqllib_cat") == None:
            self.userdb.create_table("sqllib_cat", "0.1", TABLE_LIB_CAT)
        if self.userdb.get_table_version("sqllib_sql") == None:
            self.userdb.create_table("sqllib_sql", "0.1", TABLE_LIB_SQL)

    def get_entries(self, parent=None):
        """Return rows from userdb that belong to parent (an ID).

        This yields a tuple (is_category, id, title, description, statement).
        """
        ret = []
        sql = 'select id, name, parent from sqllib_cat'
        if parent is not None:
            sql += ' where parent = %d' % parent
        else:
            sql += ' where parent is null'
        [ret.append(Category(row[0], row[1], row[2]))
         for row in self.userdb.cursor.execute(sql)]
        sql = ('select id, title, statement, description, category '
               'from sqllib_sql')
        if parent is not None:
            sql += ' where category = %d' % parent
        else:
            sql += ' where category is null'
        [ret.append(Statement(row[0], row[1], row[2], row[3], row[4]))
         for row in self.userdb.cursor.execute(sql)]
        return ret

    def save_entry(self, entry):
        """Save entry back to library."""
        if entry.id is None:
            sql, params = entry.get_insert_statement()
        else:
            sql, params = entry.get_update_statement()
        self.userdb.cursor.execute(sql, params)
        self.userdb.conn.commit()
        if entry.id is None:
            entry.id = self.userdb.cursor.lastrowid
        for instance, (view, merge_id) in self._views.iteritems():
            view.refresh()
        return entry.id

    def delete_entry(self, entry):
        """Deletes entry in library."""
        sql, params = entry.get_delete_statement()
        self.userdb.cursor.execute(sql, params)
        self.userdb.conn.commit()
        for instance, (view, merge_id) in self._views.iteritems():
            view.remove_entry(entry)


class SQLLibraryView(gtk.ScrolledWindow, PaneItem):

    name = _(u'Library')
    icon = 'stock_book_green'
    detachable = True

    def __init__(self, app, instance, library):
        """Constructor."""
        self.app = app
        self.instance = instance
        self.lib = library
        gtk.ScrolledWindow.__init__(self)
        PaneItem.__init__(self, app)
        self.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        self.model = gtk.TreeStore(gobject.TYPE_PYOBJECT,   # 0 item
                                   str,                     # 1 stock id
                                   str,                     # 2 label
                                   str,                     # 3 content
                                   bool,                    # 4 editable
                                   )
        self.list = gtk.TreeView(self.model)
        col = gtk.TreeViewColumn()
        renderer = gtk.CellRendererPixbuf()
        col.pack_start(renderer, expand=False)
        col.add_attribute(renderer, 'stock-id', 1)
        renderer = gtk.CellRendererText()
        renderer.set_property('ellipsize', pango.ELLIPSIZE_END)
        col.pack_start(renderer, expand=True)
        col.add_attribute(renderer, 'text', 2)
        col.add_attribute(renderer, 'editable', 4)
        renderer.connect('edited', self.on_entry_edited)
        self.list.append_column(col)
        self.list.set_headers_visible(False)
        self.add(self.list)
        self.list.show()
        self.list.connect("button-press-event", self.on_button_press_event)
        # Drag'n'drop
        targets = [
            ('MY_TREE_MODEL_ROW', gtk.TARGET_SAME_WIDGET, 0),
            ('text/plain', 0, 1),
        ]
        self.list.enable_model_drag_source(gtk.gdk.BUTTON1_MASK, targets,
                                           gtk.gdk.ACTION_DEFAULT|
                                           gtk.gdk.ACTION_MOVE)
        self.list.enable_model_drag_dest(targets, gtk.gdk.ACTION_DEFAULT)
        self.list.connect('drag_data_get', self.on_drag_data_get)
        self.list.connect('drag_data_received', self.on_drag_data_received)

    def _get_entry(self, is_cat, id_, iter_=None):
        """Returns path for an entry."""
        if iter_ is None:
            iter_ = self.model.get_iter_root()
        while iter_:
            obj = self.model.get_value(iter_, 0)
            if isinstance(obj, Category) and is_cat and obj.id == id_:
                return iter_
            elif isinstance(obj, Statement) and obj.id == id_:
                return iter_
            if self.model.iter_has_child(iter_):
                citer_ = self.model.iter_children(iter_)
                result = self._get_entry(is_cat, id_, citer_)
                if result is not None:
                    return result
            iter_ = self.model.iter_next(iter_)
        return None

    def _run_dblclick_action(self, treeview, event):
        x = int(event.x)
        y = int(event.y)
        time = event.time
        pthinfo = treeview.get_path_at_pos(x, y)
        if pthinfo is None:
            return
        model = treeview.get_model()
        obj = model.get_value(model.get_iter(pthinfo[0]), 0)
        if isinstance(obj, Category):
            if not treeview.row_expanded(pthinfo[0]):
                treeview.expand_row(pthinfo[0], False)
        else:
            editor = self.instance.editor_create()
            editor.set_text(obj.statement)

    def _show_popup_menu(self, treeview, event):
        x = int(event.x)
        y = int(event.y)
        time = event.time
        pthinfo = treeview.get_path_at_pos(x, y)
        if pthinfo is not None:
            model = treeview.get_model()
            obj = model.get_value(model.get_iter(pthinfo[0]), 0)
        else:
            obj = None
        menu = gtk.Menu()
        if obj is not None:
            if isinstance(obj, Category):
                item = gtk.MenuItem(_(u'Add category'))
                item.connect('activate', self.on_add_category, obj)
                item.show()
                menu.append(item)
            item = gtk.MenuItem(_(u'Rename'))
            item.connect('activate', self.on_entry_rename,
                         treeview, pthinfo[0])
            menu.append(item)
            item = gtk.MenuItem(_(u'Delete'))
            item.connect('activate', self.on_entry_delete, obj)
            item.show()
            menu.append(item)
        else:
            item = gtk.MenuItem(_(u'Add category'))
            item.connect('activate', self.on_add_category)
            item.show()
            menu.append(item)
        menu.show_all()
        menu.popup( None, None, None, event.button, time)

    # ---
    # Callbacks
    # ---

    def on_add_category(self, menuitem, parent=None):
        model = self.list.get_model()
        if parent is not None:
            iter_ = self._get_entry(True, parent.id)
        else:
            iter_ = model.append(None)
        obj = Category(None, _(u'Unknown'))
        model.set(iter_, 0, obj, 1, gtk.STOCK_OPEN,
                  2, obj.name, 3, None, 4, True)
        self.list.set_cursor(model.get_path(iter_), self.list.get_column(0),
                             True)

    def on_button_press_event(self, treeview, event):
        if event.type == gtk.gdk._2BUTTON_PRESS \
        and event.button == 1:
            self._run_dblclick_action(treeview, event)
        elif event.button == 3:
            self._show_popup_menu(treeview, event)
            return 1

    def on_drag_data_get(self, treeview, context, selection, target_id, etime):
        treeselection = treeview.get_selection()
        model, iter_ = treeselection.get_selected()
        obj = model.get_value(iter_, 0)
        if target_id == 0:  # drag'n'drop in treeview
            data = '%s::%s' % ((isinstance(obj, Category) and '1' or ''),
                               obj.id)
        elif isinstance(obj, Statement):
            data = obj.statement or obj.title
        else:
            data = ''
        selection.set(selection.target, 8, data)

    def on_drag_data_received(self, treeview, context, x, y, selection,
                              target_id, etime):
        model = treeview.get_model()
        data = selection.data
        drop_info = treeview.get_dest_row_at_pos(x, y)
        tbd = []
        if drop_info is not None:
            path, position = drop_info
            if position in [gtk.TREE_VIEW_DROP_BEFORE,
                            gtk.TREE_VIEW_DROP_AFTER]:
                return
            iter_ = model.get_iter(path)
            obj = model.get_value(iter_, 0)
            if target_id == 1:
                if (isinstance(obj, Statement)
                    and (position == gtk.TREE_VIEW_DROP_INTO_OR_AFTER
                         or position == gtk.TREE_VIEW_DROP_INTO_OR_BEFORE)):
                    obj.statement = data
                    tbd.append(obj)
                elif (position == gtk.TREE_VIEW_DROP_INTO_OR_AFTER
                      or position == gtk.TREE_VIEW_DROP_INTO_OR_BEFORE):
                    tbd.append(
                        Statement(None, data, data, category=obj.id))
            else:
                x1, x2 = data.split('::', 1)
                oiter = self._get_entry(bool(x1), int(x2))
                other_obj = model.get_value(oiter, 0)
                if (isinstance(obj, Category)
                    and (position == gtk.TREE_VIEW_DROP_INTO_OR_AFTER
                         or position == gtk.TREE_VIEW_DROP_INTO_OR_BEFORE)):
                    if isinstance(other_obj, Statement):
                        other_obj.category = obj.id
                    else:
                        other_obj.parent = obj.id
                    tbd.append(other_obj)
        elif target_id == 1:  # drop from editor
            data = '\n'.join(data.splitlines())
            stmt = Statement(None, data, data)
            tbd.append(stmt)
        if context.action == gtk.gdk.ACTION_MOVE:
            context.finish(True, True, etime)
        for item in tbd:
            self.lib.save_entry(item)
        return

    def on_entry_delete(self, menuitem, entry):
        self.lib.delete_entry(entry)

    def on_entry_edited(self, renderer, path, new_text):
        model = self.list.get_model()
        model.set_value(model.get_iter(path), 2, new_text)
        entry = model.get_value(model.get_iter(path), 0)
        if isinstance(entry, Category):
            entry.name = new_text
        else:
            entry.title = new_text
        entry.id = self.lib.save_entry(entry)

    def on_entry_rename(self, menuitem, treeview, path):
        treeview.set_cursor(path, treeview.get_column(0), True)

    # ---
    # Public methods
    # ---

    def get_selected_item(self):
        """Returns selected item or None."""
        sel = self.list.get_selection()
        model, iter_ = sel.get_selected()
        if iter_ is not None:
            return model.get_value(iter_, 0)
        return None

    def refresh(self, parent=None):
        """Update all entries in the TreeView."""
        for item in self.lib.get_entries(parent):
            is_cat = isinstance(item, Category)
            iter_ = self._get_entry(is_cat, item.id)
            if iter_ is None:
                if isinstance(item, Category):
                    parent = item.parent
                else:
                    parent = item.category
                if parent:
                    iter_ = self.model.append(self._get_entry(True, parent))
                else:
                    iter_ = self.model.append(None)
            if isinstance(item, Category):
                lbl = item.name
                content = None
                ico = gtk.STOCK_OPEN
            else:
                lbl = item.title or item.statement.split('\n')[0]
                content = item.statement
                ico = gtk.STOCK_FILE
            self.model.set_value(iter_, 0, item)
            self.model.set_value(iter_, 1, ico)
            self.model.set_value(iter_, 2, lbl)
            self.model.set_value(iter_, 3, content)
            self.model.set_value(iter_, 4, True)
            if is_cat:
                self.refresh(item.id)

    def remove_entry(self, entry):
        """Removes an entry form the list."""
        iter_ = self._get_entry(isinstance(entry, Category), entry.id)
        self.list.get_model().remove(iter_)


UI = '''
<menubar name="MenuBar">
  <menu name="Query" action="query-menu-action">
    <placeholder name="query-extensions">
      <separator />
      <menuitem name="LibSave" action="lib-save" />
    </placeholder>
  </menu>
</menubar>
'''


TABLE_LIB_CAT = """create table sqllib_cat (
    id integer primary key not null,
    name text,
    parent integer references sqllib_cat
);"""


TABLE_LIB_SQL = """create table sqllib_sql (
    id integer primary key not null,
    title text not null,
    description text,
    statement text not null,
    category integer
);"""



# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2009 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Database access."""

import binascii
import logging
import random
import os
import sys
import thread
import time
from ConfigParser import ConfigParser
from gettext import gettext as _
from threading import Thread

import gobject
import gtk

try:
    import gnomekeyring
    USE_KEYRING = True
    if (not gnomekeyring.is_available()
        or not gnomekeyring.get_default_keyring_sync()):
        USE_KEYRING = False
except ImportError:
    USE_KEYRING = False
logging.debug('Using Gnome keyring: %r' % USE_KEYRING)

import sqlparse


TRANSACTION_IDLE = 1 << 1
TRANSACTION_COMMIT_ENABLED = 1 << 2
TRANSACTION_ROLLBACK_ENABLED = 1 << 3
TRANSACTION_ACTIVE = TRANSACTION_COMMIT_ENABLED|TRANSACTION_ROLLBACK_ENABLED


# See http://www.sqlalchemy.org/docs/05/dbengine.html#supported-dbapis
DIALECTS = {
    'postgres': {
        'name': 'Postgres',
        'description': _(u'Provides access to PostgreSQL databases'),
        'dependencies': ['psycopg2'],
    },
    'sqlite': {
        'name': 'SQLite',
        'description': _(u'Provides access to SQLite databases'),
        'dependencies': [],  # we need sqlite3, but we need Python 2.5 too ;-)
    },
    'mysql': {
        'name': 'MySQL',
        'description': _(u'Provides access to MySQL databases'),
        'dependencies': ['mysqldb'],
    },
    'oracle': {
        'name': 'Oracle',
        'description': _(u'Provides access to Oracle databases'),
        'dependencies': ['cx_Oracle'],
    },
    'mssql': {
        'name': 'SQL Server',
        'description': _(u'Provides access to SQL Server databases'),
        'dependencies': ['pyodbc pymssql adodbapi'],  # alternatives
    },
#    'access': {
#        'name': 'MS Access',
#        'description': _(u'Provides access to MS Access files'),
#        'dependencies': ['win32com', 'pythoncom', 'pyodbc'],
#   },
    'firebird': {
        'name': 'Firebird',
        'description': _(u'Provides access to Firebird databses'),
        'dependencies': ['kinterbasdb'],
    },
    'informix': {
        'name': 'Informix',
        'description': _(u'Provides access to Informix databases'),
        'dependencies': ['informixdb'],
    },
    'maxdb': {
        'name': 'MaxDB',
        'description': _(u'Provides access to MaxDB databases'),
        'dependencies': ['sapdb'],
    },
}


import cf
from cf.db import backends
from cf.db.meta import DatabaseMeta
from cf.db.url import make_url
from cf.ui import dialogs
from cf.utils import Emit


class DummyDBAPIError(Exception):
    """Fake exception for backends that don't provider OperationalError"""


def availability():
    """Checks for availability of backends.

    Returns a dictionary with backend names as keys. The values are 2-tuples
    (available, reason) where *available* is a flag indicating if the
    backend is available on this system and *reason* is a message why a
    backend isn't available or ``None`` if the backend is available.
    """
    result = {}
    for name in DIALECTS:
        backend = get_dialect_backend(name)
        try:
            api = backend.dbapi()
            result[name] = (True, None)
        except ImportError, err:
            logging.debug('Dialect not available: %s', err)
            result[name] = (False, u'%s' % err)
    return result


def get_dialect_backend(dialect):
    """Return backend class for dialect."""
    modname = 'cf.db.backends.%s' % dialect
    mod = __import__(modname, fromlist=modname.split('.'))
    return mod.DRIVER


class Datasource(gobject.GObject):

    __gsignals__ = {
        'executed': (
            gobject.SIGNAL_RUN_LAST,
            gobject.TYPE_NONE,
            (gobject.TYPE_PYOBJECT,)),
        }

    def __init__(self, manager):
        self.__gobject_init__()
        self._engine = None
        self.connections = set()
        self.manager = manager
        self.internal_connection = None
        self._meta = None
        self._conn_count = 0
        self._url = None
        self._backend = None
        self.ask_for_password = False
        self.startup_commands = None
        self.name = None
        self.description = None
        self.color = None
        self.id = None

    @property
    def meta(self):
        if self._meta is None:
            if not self.connections:
                return None
            self._meta = DatabaseMeta(self)
        return self._meta

    @property
    def backend(self):
        return self._backend

    def _get_url(self):
        return self._url

    def _set_url(self, url):
        old_url = self._url
        self._url = url
        if ((old_url is not None
             and old_url.drivername != self._url.drivername)
            or old_url is None):
            self._backend = get_dialect_backend(self._url.drivername)()

    url = property(fget=_get_url, fset=_set_url)

    @property
    def public_url(self):
        """URL without password."""
        pwd = self.url.password
        if pwd is not None:
            self.url.password = 'XXX'
        ret = str(self.url)
        if pwd is not None:
            self.url.password = pwd
        return ret

    def _open_connection(self):
        """Internal method to create a Connection instance.

        This method is used for the 'Test Connection' feature.
        """
        if self.ask_for_password and self.url.password is None:
            msg = 'Please enter the password to connect to %(name)s.'
            msg = msg % {'name': self.name or str(self.url)}
            pwd = dialogs.password(_(u'Password required'), msg)
            self.url.password = pwd  # remember while this instance lives
        conn = self.backend.get_connection(self.url)
        conn = Connection(self, conn)
        self.backend.prepare_connection(conn)
        if self.startup_commands:
            conn.execute(self.startup_commands)
        return conn

    def on_connection_closed(self, connection):
        if connection in self.connections:
            self.connections.remove(connection)
        if connection == self.internal_connection:
            self.internal_connection = None
            for conn in self.connections:
                self.internal_connection = conn
                break
        self.manager.emit('datasource-changed', self)

    def dbconnect(self):
        """Returns a database connection."""
        try:
            conn = self._open_connection()
            self._conn_count += 1
            conn.num = self._conn_count
            self.connections.add(conn)
            conn.connect('closed', self.on_connection_closed)
            if self.internal_connection is None:
                self.internal_connection = conn
            if self.manager is not None:
                self.manager.emit('datasource-changed', self)
            return conn
        except:
            logging.exception('Datasource.dbconnect failed:')
            msg = _(u'Could not connect to %(name)s: %(error)s')
            msg = msg % {'name': self.get_label(),
                         'error': str(sys.exc_info()[1])}
            dialogs.error(_(u'Failed'), msg)
            return None

    def dbdisconnect_all(self):
        """Disconnect all open connections."""
        while self.connections:
            conn = self.connections.pop()
            conn.close()

    def get_connections(self):
        return self.connections

    def get_label(self):
        if self.name:
            return '%s (%s)' % (self.name, self.public_url)
        return self.public_url

    def to_dict(self):
        data = {}
        data['ask_for_password'] = self.ask_for_password
        data['name'] = self.name
        data['description'] = self.description
        data['color'] = self.color
        data['id'] = self.id
        data['url'] = self.url
        for key in ('username', 'password', 'host', 'port', 'database'):
            data[key] = getattr(self.url, key)
        data.update(self.url.query)
        return data


class Connection(gobject.GObject):

    __gsignals__ = {
        'closed': (
            gobject.SIGNAL_RUN_LAST,
            gobject.TYPE_NONE,
            tuple()),
        'notice' : (
            gobject.SIGNAL_RUN_LAST,
            gobject.TYPE_NONE,
            (str,)),
    }

    __gproperties__ = {
        'transaction-state': (
            gobject.TYPE_PYOBJECT,
            'Transaction flag', 'Transaction flag',
            gobject.PARAM_READWRITE),
    }

    def __init__(self, datasource, real_conn):
        self._transaction_state = TRANSACTION_IDLE
        self.__gobject_init__()
        self.datasource = datasource
        self.connection = real_conn
        self.num = 0

    @property
    def threadsafety(self):
        # Workaround to simulate some DB-API2 stuff
        if self.datasource.url.drivername != 'sqlite':
            return 2
        return 1

    @property
    def meta(self):
        return self.datasource.meta

    def do_set_property(self, property, value):
        if property.name == "transaction-state":
            self._transaction_state = value
        else:
            raise AttributeError, "unknown property %r" % property.name

    def do_get_property(self, property):
        if property.name == "transaction-state":
            return self._transaction_state
        else:
            raise AttributeError, "unknown property %r" % property.name

    def get_label(self, short=False):
        """Return a label for this connection.

        :param short: If ``True``, data source information will be omitted.
          Defaults to ``False``.
        """
        if short:
            return _(u'Connection #%(num)d') % {'num': self.num}
        else:
            return '%s #%d' % (self.datasource.get_label(), self.num)

    def execute(self, sql):
        cur = self.connection.cursor()
        cur.execute(sql)
        if cur.description:
            return cur.fetchall()

    def get_dbapi_connection(self):
        return self.connection

    def close(self):
        self.connection.close()
        self.emit('closed')

    def prepare_statement(self, sql):
        """Prepare sql before executing."""
        return self.datasource.backend.prepare_statement(sql)

    def execute_raw(self, sql):
        """Just for internal use!."""
        conn = self.connection
        cur = conn.cursor()
        cur.execute(sql)
        description = cur.description
        ret = []
        if description:
            for row in cur.fetchall():
                data = {}
                for idx, item in enumerate(description):
                    data[item[0]] = row[idx]
                    data[idx] = row[idx]
                ret.append(data)
            return ret

    def begin(self):
        """Start transaction."""
        # We don't support nested transactions.
        if not self.datasource.backend.begin(self):
            self.connection.begin()
        self.update_transaction_state()

    def commit(self):
        """Commit changes."""
        if not self.datasource.backend.commit(self):
            self.connection.commit()
        self.update_transaction_state()

    def rollback(self):
        """Rollback changes."""
        if not self.datasource.backend.rollback(self):
            self.connection.rollback()
        self.update_transaction_state()

    def update_transaction_state(self):
        """Update the current transaction state."""
        old_state = self.get_property('transaction-state')
        new_state = self.datasource.backend.get_transaction_state(self)
        if new_state is None:
            if self.connection.in_transaction:
                new_state = TRANSACTION_COMMIT_ENABLED | \
                            TRANSACTION_ROLLBACK_ENABLED
            else:
                new_state = TRANSACTION_IDLE
        if new_state != old_state:
            self.set_property('transaction-state', new_state)

    def explain_statements(self, statement):
        """Return a list of statements to execute for EXPLAIN."""
        explain = self.datasource.backend.get_explain_statement(statement)
        if explain is None:
            return []
        if not isinstance(explain, (list, tuple)):
            explain = [explain]
        return explain


class DatasourceManager(gobject.GObject):

    __gsignals__ = {
        'datasource-changed': (
            gobject.SIGNAL_RUN_LAST,
            gobject.TYPE_NONE,
            (gobject.TYPE_PYOBJECT,)),
        'datasource-added': (
            gobject.SIGNAL_RUN_LAST,
            gobject.TYPE_NONE,
            (gobject.TYPE_PYOBJECT,)),
        'datasource-deleted': (
            gobject.SIGNAL_RUN_LAST,
            gobject.TYPE_NONE,
            (gobject.TYPE_PYOBJECT,)),
    }

    def __init__(self, app):
        self.app = app
        self._cache = {}
        self.__gobject_init__()

    def _get_config(self):
        conf = ConfigParser()
        fname = os.path.join(cf.USER_CONFIG_DIR, 'datasources.cfg')
        if os.path.exists(fname):
            conf.readfp(open(fname))
        return conf

    def _get_id(self):
        key = ''.join(map(chr, (random.randrange(256) for i in xrange(16))))
        return binascii.hexlify(key)

    def save(self, datasource):
        """Save data source."""
        if datasource.id is None:
            datasource.id = self._get_id()
            sig_name = 'datasource-added'
        else:
            sig_name = 'datasource-changed'
            old_ds = self.load(datasource.id, cached=False)
            if old_ds.url.get_dict() != datasource.url.get_dict():
                datasource.dbdisconnect_all()
        conf = self._get_config()
        if conf.has_section(datasource.id):
            conf.remove_section(datasource.id)
        conf.add_section(datasource.id)
        pwd = None
        if (datasource.url.password and datasource.ask_for_password
            or USE_KEYRING):
            if USE_KEYRING:
                self.store_password(datasource)
            pwd = datasource.url.password
            datasource.url.password = None
        conf.set(datasource.id, 'url', datasource.url)
        conf.set(datasource.id, 'ask_for_password',
                 datasource.ask_for_password)
        conf.set(datasource.id, 'startup_commands',
                 datasource.startup_commands or '')
        conf.set(datasource.id, 'name', datasource.name or '')
        conf.set(datasource.id, 'description', datasource.description or '')
        conf.set(datasource.id, 'color', datasource.color or '')
        fp = open(os.path.join(cf.USER_CONFIG_DIR, 'datasources.cfg'), 'w')
        conf.write(fp)
        fp.close()
        if pwd is not None:
            datasource.url.password = pwd
        self._cache[datasource.id] = datasource
        self.emit(sig_name, datasource)

    def load(self, id_, conf=None, cached=True):
        """Load data source with given ID."""
        if cached and id_ in self._cache:
            return self._cache[id_]
        if conf is None:
            conf = self._get_config()
        if not conf.has_section(id_):
            raise RuntimeError('Unknown data source %s' % id_)
        ds = Datasource(manager=self)
        ds.id = id_
        ds.url = make_url(conf.get(id_, 'url'))
        self.get_password_from_keyring(ds)
        ds.ask_for_password = conf.getboolean(id_, 'ask_for_password')
        ds.startup_commands = conf.get(id_, 'startup_commands') or None
        ds.name = conf.get(id_, 'name') or None
        ds.description = conf.get(id_, 'description') or None
        ds.color = conf.get(id_, 'color') or None
        self._cache[id_] = ds
        return ds

    def delete(self, datasource):
        """Delete a data source."""
        conf = self._get_config()
        if conf.has_section(datasource.id):
            conf.remove_section(datasource.id)
        fp = open(os.path.join(cf.USER_CONFIG_DIR, 'datasources.cfg'), 'w')
        conf.write(fp)
        fp.close()
        if USE_KEYRING:
            datasource.password = None
            self.store_password(datasource)
        self.emit('datasource-deleted', datasource)

    def get_all(self):
        """Return all datasources."""
        conf = self._get_config()
        res = []
        for id_ in conf.sections():
            res.append(self.load(id_, conf))
        return res

    def get_connections(self):
        """Return a list of open connections."""
        res = []
        for datasource in self.get_all():
            res += datasource.get_connections()
        return res

    def store_password(self, datasource):
        """Update password.

        If the gnomekeyring bindings are available, the password is stored
        in the keyring. Otherwise it's stored in datasources.cfg.

        It's assumed, that the data source has an ID. So, this method
        *must* be called after the data source has been saved once.

        :param datasource: A data source.
        """
        if not USE_KEYRING:
            return

        item_type = gnomekeyring.ITEM_GENERIC_SECRET
        attrs = {"crunchyfrog": datasource.id}
        try:
            entry = gnomekeyring.find_items_sync(item_type, attrs)
            entry = entry[0]
        except gnomekeyring.NoMatchError:
            entry = None
        if datasource.url.password is not None:
            gnomekeyring.item_create_sync(None, item_type,
                                          'crunchyfrog/%s' % datasource.id,
                                          attrs, datasource.url.password, True)
        elif entry is not None:
            gnomekeyring.item_delete_sync(None, entry.item_id)

    def get_password_from_keyring(self, datasource):
        """Retrieve password from keyring and set it to datasource."""
        if not USE_KEYRING or datasource.url.password:
            return
        item_type = gnomekeyring.ITEM_GENERIC_SECRET
        attrs = {"crunchyfrog": datasource.id}
        try:
            entry = gnomekeyring.find_items_sync(item_type, attrs)
            entry = entry[0]
        except gnomekeyring.NoMatchError:
            entry = None
        if entry is not None:
            datasource.url.password = entry.secret


class Query(gobject.GObject):
    """Object representing a database query."""

    __gsignals__ = {
        "started" : (gobject.SIGNAL_RUN_LAST,
                     gobject.TYPE_NONE,
                     tuple()),
        "finished" : (gobject.SIGNAL_RUN_LAST,
                      gobject.TYPE_NONE,
                      tuple())
    }

    def __init__(self, statement, connection):
        """Constructor.

        :param statement: Raw SQL statement.
        :param connection: A database connection.
        """
        self.__gobject_init__()
        self.connection = connection
        self.statement = self.connection.prepare_statement(statement)
        self._parsed = None
        self.description = None
        self.rowcount = -1
        self.rows = None
        self.messages = []
        self.failed = False
        self.executed = False
        self.execution_time = None
        self.coding_hint = "utf-8"
        self.errors = list()
        self.error_position = None

    @property
    def parsed(self):
        if self._parsed is None:
            self._parsed = sqlparse.parse(self.statement)[0]
        return self._parsed

    def execute(self, threaded=False):
        """Execute the statement.

        :param threaded: If ``True`` the statement is executed in threaded
          mode, otherwise in blocking mode (default: ``False``).
        """
        backend = self.connection.datasource.backend
        if threaded:
            Emit(self, "started")
        else:
            self.emit("started")
        start = time.time()
        dbapi_conn = self.connection.get_dbapi_connection()
        dbapi_cur = dbapi_conn.cursor()
        operational_error = getattr(backend.dbapi(), 'OperationalError',
                                    DummyDBAPIError)
        programming_error = getattr(backend.dbapi(), 'ProgrammingError',
                                    DummyDBAPIError)
        do_close = False
        try:
            dbapi_cur.execute(self.statement)
        except Exception, err:
            self.failed = True
            self.errors.append(str(err))
            try:
                self.error_position = backend.get_error_position(self, err)
            except:
                logging.exception('Failed to get error position:')
                self.error_position = None
            do_close = backend.should_close(err)
        except:
            self.failed = True
            self.errors.append(str(sys.exc_info()[1]))
        self.executed = True
        self.execution_time = time.time() - start
        if not self.failed:
            if hasattr(dbapi_cur, 'statusmessage'):
                self.messages = [dbapi_cur.statusmessage]
            else:
                self.messages = []
            self.description = dbapi_cur.description
            self.rowcount = dbapi_cur.rowcount
            if self.description:
                self.rows = dbapi_cur.fetchall()
        self.connection.update_transaction_state()
        if threaded:
            Emit(self, "finished")
            Emit(self.connection.datasource, 'executed', self)
        else:
            self.emit("finished")
            self.connection.datasource.emit('executed', self)
        if do_close:
            logging.debug('Closing connection')
            gobject.idle_add(self.connection.close)



# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Export filter"""

import datetime
import os
import pwd
import types

from cf.plugins.core import ExportPlugin

from gettext import gettext as _

class CSVExportFilter(ExportPlugin):
    id = "crunchyfrog.export.csv"
    name = _(u"CSV Export")
    description = _(u"Export query results as CSV files")
    author = "Andi Albrecht"
    license = "GPL"
    homepage = "http://cf.andialbrecht.de"
    version = "0.1"
    has_options = False

    file_filter_name = _(u"Text CSV (.csv)")
    file_filter_pattern = ["*.csv"]

    def __init__(self, app):
        ExportPlugin.__init__(self, app)

    def export(self, description, rows, options):
        import csv
        fp = open(options["filename"], "w")
        w = csv.writer(fp)
        w.writerow([x[0] for x in description])
        w.writerows(rows)
        fp.close()

class OOCalcExportFilter(ExportPlugin):
    id = "crunchyfrog.export.odc"
    name = _(u"OpenDocument Export")
    description = _(u"Export query results as OpenDocument Spreadsheets")
    author = "Andi Albrecht"
    license = "GPL"
    homepage = "http://cf.andialbrecht.de"
    version = "0.1"

    file_filter_name = _(u"OpenDocument Spreadsheet (.ods)")
    file_filter_pattern = ["*.ods"]

    def __init__(self, app):
        ExportPlugin.__init__(self, app)

    def export(self, description, rows, options):
        import ooolib
        doc = ooolib.Calc()
        doc.set_meta('title', _(u"CrunchyFrog Data Export"))
        doc.set_meta('description', options.get("query", ""))
        doc.set_meta('creator', pwd.getpwuid(os.getuid())[4].split(",", 1)[0])
        doc.set_cell_property('bold', True)
        for i in range(len(description)):
            doc.set_cell_value(i+1, 1, "string", description[i][0])
        doc.set_cell_property('bold', False)
        for i in range(len(rows)):
            for j in range(len(rows[i])):
                value = rows[i][j]
                if value == None:
                    continue
                if type(value) == types.FloatType:
                    otype = "float"
                elif type(value) == types.IntType:
                    otype = "int"
                elif type(value) == types.BooleanType:
                    otype = "boolean"
                else:
                    otype = "string"
                doc.set_cell_value(j+1, i+2, otype, value)
        doc.save(options["filename"])


class XlsExportFilter(ExportPlugin):
    id = "crunchyfrog.export.xls"
    name = _(u"Microsoft Excel Export")
    description = _(u"Export query results as Microsoft Excel Spreadsheets")
    author = "Michele Albanese"
    license = "GPL"
    homepage = ""
    version = "0.1"

    file_filter_name = _(u"Excel Spreadsheet (.xls)")
    file_filter_pattern = ["*.xls"]

    def __init__(self, app):
        ExportPlugin.__init__(self, app)

    def export(self, description, rows, options):
        from xlwt import Workbook, easyxf
        doc = Workbook(encoding='utf-8')
	sheet = doc.add_sheet('Sheet 1', cell_overwrite_ok=True)
        header_format = easyxf('font: bold True')
        for i in range(len(description)):
            sheet.write(0, i, description[i][0], header_format)
        date_format = easyxf(num_format_str='YYYY-MM-DD')
        for i in range(len(rows)):
            for j in range(len(rows[i])):
               value = rows[i][j]
	       if isinstance(value, datetime.datetime):
		   sheet.row(i+1).set_cell_date(j, value, date_format)
	       else:
	           sheet.row(i+1).write(j, value)
	       if((i % 500) == 0):
		   sheet.flush_row_data
        doc.save(options["filename"])




# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Object browser"""

import logging
from gettext import gettext as _

import gtk
import gobject

try:
    import sexy
    HAVE_SEXY = True
except ImportError:
    HAVE_SEXY = False

from cf.db import Datasource, objects
from cf.ui.editor import SQLView
from cf.ui import pane, dialogs

import sqlparse.tokens


class DummyNode(object):
    """Placeholder object that is replaced when a node is expanded."""


class Browser(gtk.ScrolledWindow, pane.PaneItem):
    """The object browser (aka navigator) class."""

    __gsignals__ = {
        "object-menu-popup" : (
            gobject.SIGNAL_RUN_LAST,
            gobject.TYPE_NONE,
            (gobject.TYPE_PYOBJECT,  # popup menu
             gobject.TYPE_PYOBJECT,  # selected object or None
             gobject.TYPE_PYOBJECT,  # selected path or None
             )
            ),
    }

    item_id = 'navigator'
    name = _(u'Navigator')
    icon = gtk.STOCK_CONNECT
    detachable = True

    def __init__(self, app, instance):
        """Constructor."""
        pane.PaneItem.__init__(self, app)
        self.app = app
        self.instance = instance
        self._setup_widget()
        self._setup_connections()
        self.reset_tree()
        self.guess_hint()
        self.connect('object-menu-popup', self.on_object_menu_popup)

    def _setup_widget(self):
        gtk.ScrolledWindow.__init__(self)
        self.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        self.set_size_request(250, -1)
        # Model
        self.model = gtk.TreeStore(
            gobject.TYPE_PYOBJECT, # 0 Object
            str,                   # 1 label
            gtk.gdk.Pixbuf,        # 2 icon
            gobject.TYPE_PYOBJECT, # 3 double-click callback
            gobject.TYPE_PYOBJECT, # 4 popup menu callback
            str,                   # 5 tooltip
            gtk.gdk.Color,         # 6 color
            bool,                  # 7 color cell visible
            )
        self.model.set_sort_column_id(1, gtk.SORT_ASCENDING)
        # TreeView
        self.object_tree = gtk.TreeView(self.model)
        self.add(self.object_tree)
        self.object_tree.set_tooltip_column(5)
        col = gtk.TreeViewColumn('', gtk.CellRendererText(),
                                 background_gdk=6, visible=7)
        self.object_tree.append_column(col)
        col = gtk.TreeViewColumn()
        renderer = gtk.CellRendererPixbuf()
        col.pack_start(renderer, expand=False)
        col.add_attribute(renderer, 'pixbuf', 2)
        renderer = gtk.CellRendererText()
        col.pack_start(renderer, expand=True)
        col.add_attribute(renderer, 'text', 1)
        self.object_tree.append_column(col)
        self.object_tree.set_expander_column(col)
        self.object_tree.set_headers_visible(False)
        self.object_tree.enable_model_drag_source(
            gtk.gdk.BUTTON1_MASK,
            (("text/plain", gtk.TARGET_SAME_APP, 80),),
            gtk.gdk.ACTION_DEFAULT | gtk.gdk.ACTION_COPY
            )

        def drag_data_get(treeview, context, selection, info, timestamp):
            treeselection = treeview.get_selection()
            model, iter = treeselection.get_selected()
            obj = model.get_value(iter, 0)
            # row_draggable(path) doesn't work somehow...
            if isinstance(obj, objects.DBObject) \
            and not isinstance(obj, objects.Collection):
                if hasattr(obj, 'schema'):
                    # If we're directly associated with a schema, include it as a prefix
                    text = '%s.%s' % (obj.schema.name, 
                                      model.get_value(iter, 1))
                else:
                    text = model.get_value(iter, 1)
            else:
                text = ""
            selection.set('text/plain', 8, text)
            return

        self.object_tree.connect("drag_data_get", drag_data_get)
        self.set_shadow_type(gtk.SHADOW_ETCHED_IN)
        # Hint
        style = self.object_tree.get_style()
        if HAVE_SEXY:
            lbl = sexy.UrlLabel(_(u'<b>No active data sources</b>\n\n'
                                  u'<a href="#">Click</a> to open the data '
                                  u'source manager.'))
            lbl.connect("url-activated", self.instance.on_datasource_manager)
        else:
            lbl = gtk.Label(_(u'<b>No active data sources</b>\n'
                              u'Select "Edit" &gt; "Data Source Manager"\n'
                              u'to create one.'))
            lbl.set_use_markup(True)
        lbl.set_alignment(0.5, 0)
        lbl.set_padding(15, 15)
        self._hint = gtk.EventBox()
        self._hint.add(lbl)
        self._hint.modify_bg(gtk.STATE_NORMAL, style.white)

    def _setup_connections(self):
        self.app.datasources.connect('datasource-added',
                                     self.on_datasource_added)
        self.app.datasources.connect('datasource-deleted',
                                     self.on_datasource_deleted)
        self.app.datasources.connect('datasource-changed',
                                     self.on_datasource_modified)
        self.object_tree.connect('button-press-event',
                                 self.on_button_press_event)
        self.object_tree.connect('row-expanded', self.on_row_expanded)
        sel = self.object_tree.get_selection()
        sel.connect('changed', self.on_object_tree_selection_changed)

    def _create_dsinfo_menu(self, model, iter, popup):

        def connect(item, datasource_info):
            conn = datasource_info.dbconnect()
            if conn:
                self.on_object_tree_selection_changed(
                    self.object_tree.get_selection())

        def disconnect(item, datasource_info):
            datasource_info.dbdisconnect()
            iter = self.get_iter_for_datasource(datasource_info)
            model = self.object_tree.get_model()
            self.object_tree.collapse_row(model.get_path(iter))

        datasource_info = model.get_value(iter, 0)
        if datasource_info.get_connections():
            lbl = _(u"Disconnect")
            cb = disconnect
        else:
            lbl = _(u"Connect")
            cb = connect
        item = gtk.MenuItem(lbl)
        item.connect("activate", cb, datasource_info)
        popup.append(item)
        return popup

    def on_object_menu_popup(self, navigator, popup, obj, path):
        model = self.object_tree.get_model()
        if obj is None:
            item = gtk.MenuItem(_(u'Add Data source'))
            item.connect('activate', self.instance.on_datasource_manager)
            item.show()
            popup.append(item)
        elif isinstance(obj, Datasource):
            if obj.connections:
                item = gtk.MenuItem(_(u'Disconnect'))
                item.connect('activate',
                             lambda m, d: d.dbdisconnect_all(), obj)
            else:
                item = gtk.MenuItem(_(u'Connect'))
                item.connect('activate', lambda m, d: d.dbconnect(), obj)
            item.show()
            popup.append(item)
        else:
            item = gtk.MenuItem(_(u'Refresh'))
            ds = self.find_datasource_info(model, model.get_iter(path))
            conn = ds.internal_connection
            item.connect('activate',
                         lambda m, d, o: d.backend.refresh(o, d.meta, conn),
                         ds, obj)
            item.show()
            popup.append(item)

    def guess_hint(self):
        if self.object_tree.get_model().get_iter_first():
            if not self.get_child() == self.object_tree:
                self.remove(self._hint.get_parent())
                self.add(self.object_tree)
        else:
            if not self.get_child() == self._hint:
                self.remove(self.object_tree)
                self.add_with_viewport(self._hint)
        self.show_all()

    def _show_popup_menu(self, treeview, event):
        """Build and show the popup menu on RMB."""
        x = int(event.x)
        y = int(event.y)
        time = event.time
        pthinfo = treeview.get_path_at_pos(x, y)
        path = None
        popup = gtk.Menu()
        obj = None
        if pthinfo is not None:
            path, col, cellx, celly = pthinfo
            treeview.grab_focus()
            treeview.set_cursor( path, col, 0)
            model = treeview.get_model()
            iter_ = model.get_iter(path)
            obj = model.get_value(iter_, 0)
        self.emit('object-menu-popup', popup, obj, path)
        if popup.get_children():
            popup.show_all()
            popup.popup( None, None, None, event.button, time)

    def _run_dblclick_action(self, treeview, event):
        """Determine an action for a doubleclick on a node and run it."""
        x = int(event.x)
        y = int(event.y)
        time = event.time
        pthinfo = treeview.get_path_at_pos(x, y)
        if pthinfo is not None:
            path, col, cellx, celly = pthinfo
            treeview.grab_focus()
            treeview.set_cursor( path, col, 0)
            model = treeview.get_model()
            iter_ = model.get_iter(path)
            obj = model.get_value(iter_, 0)
            if isinstance(obj, Datasource) \
            and not obj.internal_connection:
                self.instance.statusbar.push(0, _(u"Connecting..."))
                conn = obj.dbconnect()
                if conn is None:
                    self.instance.statusbar.pop(0)
                    return
                editor = self.instance.editor_create()
                editor.set_connection(conn)
                self.on_object_tree_selection_changed(
                    self.object_tree.get_selection())
            elif isinstance(obj, Datasource):
                editor = self.instance.editor_create()
                editor.set_connection(obj.internal_connection)
            #else:
            #    self.on_show_details(None, obj, model, iter_)

    def on_button_press_event(self, treeview, event):
        if event.type == gtk.gdk._2BUTTON_PRESS \
        and event.button == 1:
            self._run_dblclick_action(treeview, event)
        elif event.button == 3:
            self._show_popup_menu(treeview, event)
            return 1

    def on_datasource_added(self, manager, datasource_info):
        iter_ = self.model.append(None)
        self.set_datasource_info(iter_, datasource_info)
        self.guess_hint()

    def on_datasource_deleted(self, manager, datasource_info):
        iter_ = self.get_iter_for_datasource(datasource_info)
        if not iter_:
            return
        self.model.remove(iter_)
        self.guess_hint()

    def on_datasource_modified(self, manager, datasource_info):
        iter_ = self.get_iter_for_datasource(datasource_info)
        if not iter_:
            return
        self.set_datasource_info(iter_, datasource_info)

    def on_query_executed(self, datasource, query):
        if query.failed:
            return
        if query.parsed.tokens[0].ttype is sqlparse.tokens.Keyword.DDL:
            # FIXME: implement refresh tree
            #   Only already expanded items should be refreshed.
            #   This must happen after the datasource has updated it's
            #   internal data.
            logging.error('Not implemented: Refresh tree.')

    def on_object_tree_selection_changed(self, selection):
        model, iter_ = selection.get_selected()
        if iter_ is None:
            return
        obj = model.get_value(iter_, 0)
        if isinstance(obj, Datasource):
            if obj.internal_connection:
                server_info = obj.meta.get_server_info()
                if server_info:
                    self.instance.statusbar.push(0, server_info)
                else:
                    self.instance.statusbar.pop(0)
            else:
                self.instance.statusbar.pop(0)
        else:
            comm = model.get_value(iter_, 5) or ""
            self.instance.statusbar.push(0, comm)

    def on_refresh_node(self, menuitem, model, iter):
        citer = model.iter_children(iter)
        while citer:
            model.remove(citer)
            citer = model.iter_children(iter)
        citer = model.append(iter)
        model.set_value(citer, 0, DummyNode())
        self.object_tree.emit("row-expanded", iter, model.get_path(iter))

    def on_row_expanded(self, treeview, iter, path):
        model = treeview.get_model()
        obj = model.get_value(iter, 0)
        citer = model.iter_children(iter)
        if citer:
            cobj = model.get_value(citer, 0)
            if isinstance(cobj, DummyNode):
                datasource_info = self.find_datasource_info(model, citer)
                if datasource_info.color is not None:
                    color = gtk.gdk.color_parse(datasource_info.color)
                else:
                    color = None
                model.remove(citer)
                if datasource_info.meta is None:
                    return
                if isinstance(obj, Datasource):
                    obj = None
                icon = self.app.load_icon(gtk.STOCK_OPEN,
                                          gtk.ICON_SIZE_MENU,
                                          gtk.ICON_LOOKUP_FORCE_SVG)
                for child in datasource_info.meta.get_children(obj):
                    citer = model.append(iter)
                    model.set(citer,
                              0, child,
                              1, child.get_display_name(),
                              2, child.get_icon_pixbuf(),
                              5, child.comment,
                              6, color,
                              7, True)
                    if child.has_children():
                        cciter = model.append(citer)
                        model.set(cciter, 0, DummyNode(), 7, False)
                treeview.expand_row(model.get_path(iter), False)
                return

    def on_show_details(self, menuitem, object, model, iter):
        # FIXME(andi): Rewrite this part when main notebook allows different
        #  views.
        datasource_info = self.find_datasource_info(model, iter)
        if datasource_info.meta:
            data = datasource_info.meta.get_details(object)
            if not data:
                return
            for key, value in data.items():
                if isinstance(value, basestring):
                    editor = self.instance.editor_create()
                    editor.set_text(value)
                    return


    def find_datasource_info(self, model, iter):
        while iter:
            obj = model.get_value(iter, 0)
            if isinstance(obj, Datasource):
                return obj
            iter = model.iter_parent(iter)

    def get_iter_for_datasource(self, datasource_info):
        iter_ = self.model.get_iter_first()
        while iter_:
            if self.model.get_value(iter_, 0) == datasource_info:
                return iter_
            iter_ = self.model.iter_next(iter_)
        return None

    def reset_tree(self):
        self.model.clear()
        for datasource_info in self.app.datasources.get_all():
            iter_ = self.model.append(None)
            self.set_datasource_info(iter_, datasource_info)

    def set_datasource_info(self, iter, datasource_info):
        if datasource_info.get_connections():
            ico = self.app.load_icon(gtk.STOCK_CONNECT)
        else:
            ico = self.app.load_icon(gtk.STOCK_DISCONNECT)
        if datasource_info.color is not None:
            color = gtk.gdk.color_parse(datasource_info.color)
        else:
            color = None
        self.model.set(iter,
                  0, datasource_info,
                  1, datasource_info.get_label(),
                  2, ico,
                  3, None,
                  4, self._create_dsinfo_menu,
                  5, datasource_info.description,
                  6, color,
                  7, True)
        citer = self.model.iter_children(iter)
        if datasource_info.get_connections() and not citer:
            citer = self.model.append(iter)
            self.model.set(citer, 0, DummyNode())
        elif not datasource_info.get_connections() and citer:
            while citer:
                self.model.remove(citer)
                citer = self.model.iter_children(iter)
        # connect executed signal
        if not datasource_info.get_data('browser-sig-executed'):
            sig = datasource_info.connect('executed', self.on_query_executed)
            datasource_info.set_data('browser-sig-executed', sig)



# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Panes in the mainwindow"""

import logging
import os
import re
from gettext import gettext as _

import gobject
import gtk
import pango

try:
    # Let's see if we have additional items.
    # This is a dirty hack...
    it = gtk.icon_theme_get_default()
    info = it.load_icon('stock_book_green', gtk.ICON_SIZE_MENU,
                        gtk.ICON_LOOKUP_FORCE_SVG)
    HAVE_ICONS = True
except gobject.GError, err:
    # Some special icons are not present. Let's set HAVE_ICONS to False.
    # This causes the tab widgets to display the name of the pane item.
    # It's better than empty tab labels... ;-)
    logging.warning('Failed to load additional icons (%s)', err)
    HAVE_ICONS = False


# GTK style from GEdit's close button (gedit-notebook.c)
gtk.rc_parse_string('''style "cf-tab-button-style"
{
  GtkWidget::focus-padding = 0
  GtkWidget::focus-line-width = 0
  xthickness = 0
  ythickness = 0
}
widget "*.cf-tab-button" style "cf-tab-button-style"''')

TAB_LABEL_CENTER = 1
TAB_LABEL_SIDEPANE = 2


class PaneItem(object):
    name = None
    icon = None
    detachable = False

    def __init__(self, app):
        self._tab_label = None
        self.app = app

    @property
    def tab_label(self):
        return self.get_tab_label()

    def close(self, force=False):
        self.destroy()

    def get_tab_label(self):
        if self._tab_label is None:
            self._tab_label = TabLabelDefault(self.app, self)
        return self._tab_label

    def get_pane_icon(self):
        return gtk.STOCK_MISSING_IMAGE

    def get_pane_name(self):
        return '???'

    def get_pane_detachable(self):
        return False

    def get_focus_child(self):
        return None

    def get_widget(self):
        """Returns the real gtk.Widget that should be displayed."""
        return self


class Pane(object):
    """Base class for all pane widgets."""

    def __init__(self, mainwin):
        self.app = mainwin.app
        # TODO: Rename to self.instance
        self.mainwin = mainwin
        self._merge_ui()
        self.header = gtk.HBox()
        self.header.set_spacing(5)
        self.header.set_border_width(7)
        self.top_img = gtk.Image()
        self.header.pack_start(self.top_img, False, False)
        self.top_img.show()
        self.top_label = gtk.Label()
        self.top_label.set_alignment(0, 0.5)
        self.header.pack_start(self.top_label)
        self.top_label.show()
        self.top_btn = gtk.Button()
        self.top_btn.add(gtk.Arrow(gtk.ARROW_DOWN, gtk.SHADOW_NONE))
        self.top_btn.set_relief(gtk.RELIEF_NONE)
        self.top_btn.connect('button-press-event', self.on_top_btn_clicked)
        self.header.pack_start(self.top_btn, False, False)
        self.top_btn.show_all()
        self.pack_start(self.header, False, False)
        self.header.show()
        self.notebook = gtk.Notebook()
        self.notebook.set_tab_pos(gtk.POS_BOTTOM)
        self.notebook.set_scrollable(True)
        self.pack_start(self.notebook, True, True)
        self.notebook.show()
        self._autohide_tabs = False
        self.notebook.connect('switch-page', self.on_notebook_switch_page)
        self.notebook.connect('page-added',
                              self.on_notebook_pages_changed, 'added')
        self.notebook.connect('page-removed',
                              self.on_notebook_pages_changed, 'removed')
        self.notebook.set_group_id(99)
        self.hide()
        self._state_restored = False

    def _merge_ui(self):
        ui = self.mainwin.ui
        for group in ui.get_action_groups():
            if group.get_name() == 'instance':
                group.add_toggle_actions(self.get_action_entries())
        ui_desc = self.get_ui_description()
        if ui_desc:
            self.mainwin.ui.add_ui_from_string(ui_desc)
            self.mainwin.ui.ensure_update()

    def _restore(self):
        action_name = 'instance-toggle-%s' % self.__class__.__name__.lower()
        conf_name = 'win.%s_visible' % self.__class__.__name__.lower()
        action = self.mainwin._get_action(action_name)
        action.set_active(self.app.config.get(conf_name, True))
        self.on_toggle(action)

    def on_move_page(self, menuitem, page, pane):
        pane.add_item(page)

    def on_notebook_pages_changed(self, nb, page, page_num, action_str):
        if self._autohide_tabs:
            nb.set_show_tabs(nb.get_n_pages() > 1)
            nb.set_show_border(nb.get_n_pages() > 1)
        if not self._state_restored:  # don't do anything before state restored
            return
        action_name = 'instance-toggle-%s' % self.__class__.__name__.lower()
        action = self.mainwin._get_action(action_name)
        if nb.get_n_pages() == 0:
            self.hide()
            action.set_active(False)
            action.set_sensitive(False)
        else:
            if not action.get_sensitive():
                action.set_sensitive(True)
                if action.get_active():
                    self.show()
            visopt = 'win.%s_visible' % self.__class__.__name__.lower()
            if self.app.config.get(visopt, False) and nb.get_n_pages():
                action.set_active(True)

    def on_notebook_switch_page(self, nb, unusable, page_num):
        page = nb.get_nth_page(page_num)
        if page.get_data('cf::real-object'):
            page = page.get_data('cf::real-object')
        self.top_label.set_text(page.name)
        pb = self.app.load_icon(page.icon, gtk.ICON_SIZE_MENU,
                                gtk.ICON_LOOKUP_FORCE_SVG)
        self.top_img.set_from_pixbuf(pb)

    def on_page_detach(self, menuitem, page):
        alloc = page.get_allocation()
        def attach(window, page):
            self.add_item(page)
        w = gtk.Window()
        w.resize(alloc.width, alloc.height)
        w.set_transient_for(self.mainwin)
        w.connect('destroy', attach, page)
        page.reparent(w)
        w.set_title(page.name)
        w.set_icon_name(page.icon)
        w.show()

    def on_toggle(self, toggle_action):
        if toggle_action.get_active():
            self.show()
        else:
            self.hide()

    def on_top_btn_clicked(self, btn, event):
        menu = gtk.Menu()
        action_name = 'instance-toggle-%s' % self.__class__.__name__.lower()
        action = self.mainwin._get_action(action_name)
        item = action.create_menu_item()
        menu.append(item)
        item.show()
        page = self.notebook.get_nth_page(self.notebook.get_current_page())
        if page.detachable:
            item = gtk.MenuItem(_(u'Detach %(name)s') % {'name': page.name})
            item.connect('activate', self.on_page_detach, page)
            menu.append(item)
            item.show()
            if self.__class__.__name__.startswith('Bottom'):
                other_pane = _(u'side pane')
                other = self.mainwin.side_pane
            else:
                other_pane = _(u'bottom pane')
                other = self.mainwin.bottom_pane
            item = gtk.MenuItem(_(u'Move to %(name)s') % {'name': other_pane})
            item.connect('activate', self.on_move_page, page, other)
            item.show()
            menu.append(item)
        menu.popup(None, None, None, event.button, event.time)

    def get_action_entries(self):
        raise NotImplementedError

    def get_ui_description(self):
        return self.UI_DESCRIPTION

    def add_item(self, item):
        """Add a PaneItem."""
        assert isinstance(item, PaneItem)
        pitem = item
        item = pitem.get_widget()
        item.set_data('cf::real-object', pitem)
        tab_label = pitem.get_tab_label()
        if self.__class__.__name__ != 'CenterPane':
            tab_label.set_mode(TAB_LABEL_SIDEPANE)
        if item.get_parent():
            item.reparent(self.notebook)
            self.notebook.set_tab_label(item, tab_label)
        else:
            self.notebook.append_page(item, tab_label)
        self.notebook.set_tab_detachable(item, True)
        item.show()
        tab_label.show()

    def remove_item(self, pane_item):
        """Remove a previously added PaneItem."""
        raise NotImplementedError

    def set_active_item(self, item_id):
        """Toggle the active/visible item."""
        for i in range(self.notebook.get_n_pages()):
            page = self.notebook.get_nth_page(i)
            if page.item_id == item_id:
                self.notebook.set_current_page(i)
                return

    def set_autohide_tabs(self, autohide):
        """Enable/disable tab autohide."""
        self._autohide_tabs = autohide

    def set_hidden(self, hidden):
        """Hide/show the pane."""
        raise NotImplementedError

    def set_show_header(self, visible):
        """Show or hide the header."""
        if visible:
            self.header.show()
        else:
            self.header.hide()

    def state_restore(self):
        """Restore last UI state."""
        self._restore()
        self._state_restored = True
        self.on_notebook_pages_changed(self.notebook, None, 0, 'removed')


class SidePane(gtk.VBox, Pane):

    UI_DESCRIPTION = '''<menubar name="MenuBar">
      <menu name="View" action="view-menu-action">
        <menuitem name="SidePane" action="instance-toggle-sidepane" />
      </menu>
    </menubar>'''

    def __init__(self, mainwin):
        gtk.VBox.__init__(self)
        Pane.__init__(self, mainwin)

    def get_action_entries(self):
        return (
            ('instance-toggle-sidepane', None,
             _(u'Side _Pane'), 'F9',
             _(u'Show or hide the side pane in the current window'),
             self.on_toggle),
        )


class BottomPane(gtk.VBox, Pane):

    UI_DESCRIPTION = '''<menubar name="MenuBar">
      <menu name="View" action="view-menu-action">
        <menuitem name="BottomPane" action="instance-toggle-bottompane" />
      </menu>
    </menubar>'''

    def __init__(self, mainwin):
        gtk.VBox.__init__(self)
        Pane.__init__(self, mainwin)
        self.notebook.set_tab_pos(gtk.POS_RIGHT)

    def get_action_entries(self):
        return (
            ('instance-toggle-bottompane', None,
             _(u'_Bottom Pane'), '<control>F9',
             _(u'Show or hide the bottom pane in the current window'),
             self.on_toggle),
        )


class CenterPane(gtk.VBox, Pane):

    UI_DESCRIPTION = None

    def __init__(self, mainwin):
        gtk.VBox.__init__(self)
        Pane.__init__(self, mainwin)
        self.set_autohide_tabs(False)
        self.set_show_header(False)
        self.notebook.set_scrollable(True)
        self.notebook.set_tab_pos(gtk.POS_TOP)
        self.notebook.popup_disable()
        self.notebook.set_show_tabs(True)
        self.notebook.set_show_border(True)
        self.notebook.set_property("enable-popup", False)
        self.notebook.connect("switch-page", self.on_switch_page)
        self.notebook.connect("page-removed", self.on_page_removed)
        self.notebook.set_group_id(1)
        self.notebook.connect('create-window', self.on_create_window)

    def on_create_window(self, nb, page, x, y):
        view = self.get_view_by_page(page)
        view.detach()

    def on_page_added(self, notebook, child, page_num):
        gobject.idle_add(self.notebok.set_current_page, page_num)

    def on_page_removed(self, notebook, child, page_num):
        from cf.ui.editor import Editor  # import hook
        if isinstance(child, Editor):
            self.mainwin.set_editor_active(child, False)

    def on_switch_page(self, notebook, page, page_num):
        editor = self.get_nth_page(page_num).get_data('cf::real-object')
        gobject.idle_add(self.mainwin.set_editor_active, editor, True)

    def add_item(self, editor):
        super(CenterPane, self).add_item(editor)
        from cf.ui.editor import Editor
        if isinstance(editor, Editor):
            widget = editor.widget
            self.notebook.set_tab_label(widget, TabLabel(editor))
        else:
            widget = getattr(editor, 'widget', editor)  # LDAP hack
        self.notebook.set_tab_reorderable(widget, True)
        self.notebook.set_tab_detachable(widget, True)
        self.notebook.set_current_page(self.notebook.page_num(widget))
        self.mainwin._rebuild_activate_editor_actions()
        widget.connect('destroy', lambda *a:
                       self.mainwin._rebuild_activate_editor_actions())

    def get_all_editors(self):
        """Return a list of all editors."""
        ret = []
        for i in range(self.notebook.get_n_pages()):
            ret.append(self.get_view_by_pagenum(i))
        return ret

    def __getattribute__(self, name):
        """We want it to behave much like a gtk.Notebook."""
        try:
            attr = super(CenterPane, self).__getattribute__(name)
        except AttributeError:
            attr = getattr(self.notebook, name)
        return attr

    def get_action_entries(self):
        return []

    def get_view_by_pagenum(self, page_num):
        """Return editor object by page number."""
        child = self.notebook.get_nth_page(page_num)
        return self.get_view_by_page(child)

    def get_view_by_page(self, page):
        """Return editor object by child widget."""
        lbl = self.notebook.get_tab_label(page)
        return lbl.pane_item


class TabLabelDefault(gtk.HBox):
    """Default tab label implementation for pane items."""

    def __init__(self, app, pane_item, mode=TAB_LABEL_CENTER):
        """Constructor.

        :param pane_item: A :class:`PaneItem` instance.
        """
        gtk.HBox.__init__(self)
        self.app = app
        self.pane_item = pane_item
        self.mode = None
        self.datasource = None

        self.set_spacing(4)

        # Data source color indicator
        self.datasource_color_eb = gtk.EventBox()
        self.datasource_color_eb.set_size_request(5, -1)
        self.datasource_color_eb.show()
        self.pack_start(self.datasource_color_eb, False, False)

        # Icon
        pb = self.app.load_icon(self.pane_item.icon, gtk.ICON_SIZE_MENU,
                                gtk.ICON_LOOKUP_FORCE_SVG)
        self.img = gtk.image_new_from_pixbuf(pb)
        self.img.set_tooltip_text(self.pane_item.name)
        self.img.show()
        self.pack_start(self.img, False, False)

        # Label
        self.label = gtk.Label(_(u"Query"))
        self.label.set_ellipsize(pango.ELLIPSIZE_END)
        self.label.set_width_chars(15)
        self.label.set_single_line_mode(True)
        self.label.set_alignment(0, 0.5)
        # use a slightly smaller font in tabs like Empathy does...
        font_desc = self.label.get_style().font_desc
        font_desc.set_size(int(font_desc.get_size()*.8))
        self.label.modify_font(font_desc)
        self.label.show()
        self.pack_start(self.label, True, True)

        # Close button
        self.btn_close = gtk.Button()
        self.btn_close.set_focus_on_click(False)
        self.btn_close.set_relief(gtk.RELIEF_NONE)
        self.btn_close.set_name('cf-tab-button')
        image = gtk.image_new_from_stock(gtk.STOCK_CLOSE,
                                         gtk.ICON_SIZE_MENU)
        self.btn_close.add(image)
        self.btn_close.set_tooltip_text(_(u'Close'))
        self.btn_close.connect("clicked", self.on_button_close_clicked)
        self.btn_close.show_all()
        self.pack_start(self.btn_close, False, False)

        self.app.datasources.connect('datasource-changed',
                                     self.on_datasource_changed)

        self.set_mode(mode)

    def on_button_close_clicked(self, button):
        self.pane_item.close()

    def on_datasource_changed(self, manager, datasource):
        if datasource == self.datasource:
            self.set_datasource(datasource)

    def set_close_callback(self, callback):
        """Set a callback function that handles a click on the close button.

        The callback function has the signature ``def callback(tab_label)``.
        """
        assert (hasattr(callback, '__call__') or callback is None)
        self._close_cb = callback

    def set_datasource(self, datasource=None):
        """Set data source information.

        Usually this just updates the data source's color code.

        :param datasource: A :class:`~cf.db.Datasource` instance or ``None``.
        """
        if datasource is not None and datasource.color:
            color = gtk.gdk.color_parse(datasource.color)
            self.datasource_color_eb.modify_bg(gtk.STATE_NORMAL, color)
            self.datasource_color_eb.modify_bg(gtk.STATE_ACTIVE, color)
        else:
            self.datasource_color_eb.modify_bg(gtk.STATE_NORMAL, None)
            self.datasource_color_eb.modify_bg(gtk.STATE_ACTIVE, None)
        self.datasource = datasource

    def set_mode(self, mode):
        """Set tab label mode.

        *mode* is either ``TAB_LABEL_CENTER`` or ``TAB_LABEL_SIDEPANE``.
        """
        assert mode in (TAB_LABEL_CENTER, TAB_LABEL_SIDEPANE)
        self.mode = mode
        if self.mode == TAB_LABEL_CENTER:
            self.datasource_color_eb.show()
            self.label.show()
            self.btn_close.show()
        else:
            self.datasource_color_eb.hide()
            self.label.hide()
            self.btn_close.hide()

    def set_text(self, text):
        """Set tab label's text."""
        self.label.set_text(text)

    def set_tootltip(self, markup):
        """Set the tooltip for the whole tab label."""
        self.label.set_tooltip_markup(markup)
        self.datasource_color_eb.set_tooltip_markup(markup)

    def set_show_close_button(self, visible):
        """Wether the close button should be visible or not."""
        if visible:
            self.btn_close.show()
        else:
            self.btn_close.hide()



class TabLabel(TabLabelDefault):

    def __init__(self, editor):
        TabLabelDefault.__init__(self, editor.app, editor)
        self.editor = editor
        self.set_close_callback(lambda tl: tl.pane_item.close())
        from cf.ui.editor import Editor
        if isinstance(editor, Editor):
            self.editor.connect("connection-changed",
                                self.on_editor_connection_changed)
            buffer = self.editor.textview.get_buffer()
            buffer.connect("changed", self.on_buffer_changed)
            self.update_label(buffer)
        self.update_tooltip()
        self.show_all()

    def on_buffer_changed(self, buffer):
        gobject.idle_add(self.update_label, buffer)

    def on_close_editor(self, *args):
        self.editor.close()

    def on_editor_connection_changed(self, editor, connection):
        self.update_tooltip()

    def on_show_in_separate_window(self, item):
        gobject.idle_add(self.editor.show_in_separate_window)

    def update_label(self, buffer):
        if self.editor.get_filename():
            fname = self.editor.get_filename()
            txt = os.path.split(fname)[-1]
            self.label.set_tooltip_text(fname)
            if self.editor.file_contents_changed():
                txt = "*"+txt
        else:
            self.label.set_tooltip_text("")
            txt = buffer.get_text(*buffer.get_bounds())
            txt = re.sub("\s+", " ", txt)
        txt = txt.strip()
        if not txt:
            txt = _(u"Query")
        self.set_text(txt)
        self.update_tooltip()

    def update_tooltip(self):
        from cf.ui.editor import Editor
        if isinstance(self.editor, Editor):
            markup = "<b>Connection:</b> "
            if self.editor.connection:
                markup += self.editor.connection.get_label()
            else:
                markup += "["+_(u"Not connected")+"]"
            if self.editor.get_filename():
                markup += "\n<b>File:</b> "+self.editor.get_filename()
            self.set_tootltip(markup)
            conn = self.editor.connection
            if conn:
                self.set_datasource(conn.datasource)
            else:
                self.set_datasource(None)




# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Release information"""

name = "CrunchyFrog"
appname = name.lower()
version = "0.4.2"
copyright = "Copyright (C) 2009 Andi Albrecht <albrecht.andi@gmail.com>"
description = "SQL client (not-only) for GNOME"
author = "Andi Albrecht"
author_email = "albrecht.andi@gmail.com"
url = "http://crunchyfrog.googlecode.com"
long_description = """CrunchyFrog is a database navigator and SQL client.

It's mainly written for GNOME but it should run in other environments too.

Supported databases
===================

 * PostgreSQL
 * MySQL
 * SQLite3
 * Oracle
 * SQL Server
 * Firebird
 * Informix
 * MaxDB

It is written in Python (PyGTK) and can be extended using plugins.

Check out the homepage for the most recent version and details."""

license = """This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>."""

translators="""Thanks to all Launchpad contributors!
Alexandr Udovichenko (https://translations.launchpad.net/~gragdanin)
Anthony Guéchoum (https://launchpad.net/~athael)
cd (https://launchpad.net/~zadoev)
Christophe Chauvet (https://launchpad.net/~christophe-chauvet)
Christian Olate Opazo (https://launchpad.net/~cristian-olate)
Claudio Giordano
Daniel Nylander (https://translations.launchpad.net/~yeager)
Dante Díaz (https://launchpad.net/~dante)
David A Páez (https://launchpad.net/~orlsend)
DiegoJ (https://launchpad.net/~diegojromerolopez)
dvdmeer (https://translations.launchpad.net/~dennis-dvdmeer)
Edwin Dijk (https://launchpad.net/~sitix)
Enrico Rosina (https://launchpad.net/~e-rosina)
Feder Sáiz (https://launchpad.net/~federsaiz)
Guybrush88 (https://launchpad.net/~erpizzo)
Jen Ockwell (https://launchpad.net/~rj-ockwell)
Jeppe Toustrup (https://launchpad.net/~tenzer)
Kari Argillander (https://launchpad.net/~kari-argillander)
kezar (https://launchpad.net/~kezar)
Linus C Unneback (https://launchpad.net/~linus-folkdatorn)
Luca Livraghi (https://launchpad.net/~luca91)
Marion Cisternas Milla (https://launchpad.net/~mcisternas)
Martin Picek <picek.martin@gnome-cesko.cz>
Micke Elm (https://launchpad.net/~mickeelm)
Muhammad Zulfikar (https://launchpad.net/~zulfikars)
Nicola Piovesan (https://launchpad.net/~piovesannicola)
Penegal (https://launchpad.net/~penegal)
Pierre Slamich (https://launchpad.net/~pierre-slamich)
rcares (https://launchpad.net/~rcares)
Ryutatsu Ishigami (https://launchpad.net/~ai-inside+ubuntu)
Skiros (https://launchpad.net/~magoskiros)
Thanos Chatziathanassiou (https://launchpad.net/~tchatzi)
TRanslate (https://launchpad.net/~translatematic)
ubby (https://launchpad.net/~kostas-sytske)
Vadim Peretokin (https://launchpad.net/~vperetokin)
Xoan Sampaiño (https://launchpad.net/~xoan)
Yaron (https://launchpad.net/~sh-yaron)
0x00080h (https://launchpad.net/~einstein89)
"""



# -*- coding: utf-8 -*-

"""Sphinx builder that generates wiki pages for Google Code."""

import codecs
import sys
from os import path

from docutils import nodes, writers
from docutils.io import StringOutput

from sphinx.builder import TextBuilder
from sphinx.util import ensuredir, os_path
from sphinx.textwriter import TextTranslator


class GCWikiBuilder(TextBuilder):

    name = 'gcwiki'
    out_suffix = '.wiki'

    def __init__(self, *args, **kwds):
        TextBuilder.__init__(self, *args, **kwds)
        self.sections = []

    def get_target_uri(self, docname, typ=None):
        return docname.capitalize()

    def prepare_writing(self, docnames):
        self.writer = GCWikiWriter(self)

    def write_doc(self, docname, doctree):
        # Mostly copied from TextBuilder
        destination = StringOutput(encoding='utf-8')
        self.writer.write(doctree, destination)
        outfilename = path.join(self.outdir,
                                os_path(docname.capitalize())+self.out_suffix)
        # normally different from self.outdir
        ensuredir(path.dirname(outfilename))
        try:
            f = codecs.open(outfilename, 'w', 'utf-8')
            try:
                f.write(self.writer.output)
            finally:
                f.close()
        except (IOError, OSError), err:
            self.warn("Error writing file %s: %s" % (outfilename, err))

    def finish(self):
        docname = 'Sidebar'
        outfilename = path.join(self.outdir,
                                os_path(docname.capitalize())+self.out_suffix)
        # normally different from self.outdir
        ensuredir(path.dirname(outfilename))
        sidebar = ''
        self.info('Writing sidebar')
        for level, node in self.sections:
            sidebar += '*%s %s %s\n' % (level, str(node.list_attributes), node.astext())
        try:
            f = codecs.open(outfilename, 'w', 'utf-8')
            try:
                f.write(sidebar)
            finally:
                f.close()
        except (IOError, OSError), err:
            self.warn("Error writing file %s: %s" % (outfilename, err))


class GCWikiWriter(writers.Writer):
    supported = ('text',)
    settings_spec = ('No options here.', '', ())
    settings_defaults = {}

    output = None

    def __init__(self, builder):
        writers.Writer.__init__(self)
        self.builder = builder

    def translate(self):
        visitor = GCWikiTranslator(self.document, self.builder)
        self.document.walkabout(visitor)
        self.output = visitor.body


class GCWikiTranslator(TextTranslator):

    def __init__(self, document, builder):
        TextTranslator.__init__(self, document, builder)
        self._builder = builder

#    def visit_target(self, node):
#        self.add_text('[%s ' % node.astext())
#    def depart_target(self, node):
#        self.add_text(']')

    def visit_section(self, node):
        self.sectionlevel += 1
        if self.sectionlevel > 6:
            self.sectionlevel = 6
        self._title_char = '='*self.sectionlevel
    def depart_section(self, node):
        self._title_char = ''
        self.sectionlevel -= 1

    def depart_title(self, node):
        if isinstance(node.parent, nodes.section):
            char = self._title_char
        else:
            char = '^'
        text = ''.join(x[1] for x in self.states.pop() if x[0] == -1)
        self.stateindent.pop()
        text = '%s %s %s' % ('='*self.sectionlevel, text,
                             '='*self.sectionlevel)
        self.states[-1].append((0, ['', text, '']))
        self._builder.sections.append((self.sectionlevel, node))

    def visit_emphasis(self, node):
        self.add_text('_')
    def depart_emphasis(self, node):
        self.add_text('_')

    def visit_literal_emphasis(self, node):
        self.add_text('_')
    def depart_literal_emphasis(self, node):
        self.add_text('_')

    def visit_strong(self, node):
        self.add_text('*')
    def depart_strong(self, node):
        self.add_text('*')

    def visit_title_reference(self, node):
        self.add_text('*')
    def depart_title_reference(self, node):
        self.add_text('*')

    def visit_literal(self, node):
        self.add_text('`')
    def depart_literal(self, node):
        self.add_text('`')

    def visit_subscript(self, node):
        self.add_text(',,sub,,')
    def depart_subscript(self, node):
        pass

    def visit_superscript(self, node):
        self.add_text('^super^')
    def depart_superscript(self, node):
        pass

    def visit_literal_block(self, node):
        self.new_state(0)
        self.add_text('{{{\n')
    def depart_literal_block(self, node):
        self.add_text('\n}}}')
        self.end_state(wrap=False)

    visit_doctest_block = visit_literal_block
    depart_doctest_block = depart_literal_block
    visit_line_block = visit_literal_block
    depart_line_block = depart_literal_block

    def visit_pending_xref(self, node):
        self.add_text('Pxref')
    def depart_pending_xref(self, node):
        pass

    def visit_reference(self, node):
        # attrs: ds, classes, names, dupnames, backrefs
        #self.add_text('%r' % repr(node.asdom().toxml()))
        if node.hasattr('refuri'):
            self.add_text('[%s ' % node['refuri'])
        else:
            self.add_text('NO REFURI')
    def depart_reference(self, node):
        self.add_text(']')


def setup(app):
    app.add_builder(GCWikiBuilder)
